@@ -17,9 +17,15 @@
 #include "termemu.h"
 #endif
 
+#if	defined (_NOEXTRAATTR) || defined (NOUID)
 #define	MAXATTRSEL	2
 #define	ATTR_X		35
 #define	ATTR_Y		L_INFO
+#else
+#define	MAXATTRSEL	3
+#define	ATTR_X		0
+#define	ATTR_Y		L_HELP
+#endif
 
 extern int curcolumns;
 extern int mark;
@@ -29,7 +35,6 @@ extern int isearch;
 extern int fnameofs;
 extern int chgorder;
 extern int stackdepth;
-extern int removepolicy;
 extern namelist filestack[];
 #ifndef	_NOARCHIVE
 extern char archivedir[];
@@ -49,6 +54,7 @@ extern int fdmode;
 #endif
 #ifndef	_NOPTY
 extern int ptymode;
+extern int ptyinternal;
 #endif
 
 static VOID NEAR replacefname __P_((char *));
@@ -132,6 +138,9 @@ static int unpack_tree __P_((char *));
 static int info_filesys __P_((char *));
 static int NEAR selectattr __P_((char *));
 static int attr_file __P_((char *));
+#ifndef	_NOEXTRAATTR
+static int attr_dir __P_((char *));
+#endif
 #ifndef	_NOTREE
 static int tree_dir __P_((char *));
 #endif
@@ -263,6 +272,9 @@ bindtable bindlist[MAXBINDTABLE] = {
 	{'w',		WRITE_DIR,	255},
 #endif
 	{'x',		EXECUTE_FILE,	255},
+#ifndef	_NOEXTRAATTR
+	{'A',		ATTR_DIR,	255},
+#endif
 #ifndef	_NOTREE
 	{'C',		COPY_TREE,	255},
 #endif
@@ -313,8 +325,9 @@ char *name;
 	if (filepos < maxfile) free(filelist[filepos].name);
 	else for (maxfile = 0; maxfile < filepos; maxfile++)
 		if (!filelist[maxfile].name)
-			filelist[maxfile].name = strdup2("");
-	filelist[filepos].name = (name) ? name : strdup2("..");
+			filelist[maxfile].name = strdup2(nullstr);
+	filelist[filepos].name = (name) ? name : strdup2(parentpath);
+	filelist[filepos].tmpflags |= F_ISCHGDIR;
 }
 
 static int cur_up(arg)
@@ -517,14 +530,13 @@ char *arg;
 			if (!(filelist[filepos].linkname)) i = 0;
 			else {
 				i = strlen(filelist[filepos].linkname);
-				if (i > (int)sizeof(tmp) - 1)
-					i = (int)sizeof(tmp) - 1;
+				if (i > strsize(tmp)) i = strsize(tmp);
 				strncpy(tmp, filelist[filepos].linkname, i);
 			}
 		}
 		else
 # endif
-		i = Xreadlink(fnodospath(path, filepos), tmp, sizeof(tmp) - 1);
+		i = Xreadlink(fnodospath(path, filepos), tmp, strsize(tmp));
 		tmp[i] = '\0';
 		i = 1 - (strlen3(tmp) + 4);
 	}
@@ -737,7 +749,7 @@ char *arg;
 	else
 #endif
 	if (!isdir(&(filelist[filepos]))
-	|| !strcmp(filelist[filepos].name, "."))
+	|| isdotdir(filelist[filepos].name) == 2)
 		return(warning_bell(arg));
 
 	return(FNC_CHDIR);
@@ -890,7 +902,7 @@ char *arg;
 {
 	char *path;
 
-	path = strdup2(_SS_);
+	path = strdup2(rootpath);
 	if (chdir3(path, 1) < 0) error(path);
 	replacefname(path);
 
@@ -1065,7 +1077,8 @@ char *arg;
 	}
 #endif
 #ifndef	_NODOSDRIVE
-	else if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(FNC_CANCEL);
+	else if ((drive = tmpdosdupl(nullstr, &dir, 1)) < 0)
+		return(FNC_CANCEL);
 #endif
 
 	if (!execenv("FD_PAGER", filelist[filepos].name)) {
@@ -1108,7 +1121,7 @@ char *arg;
 	if (archivefile) return(FNC_CANCEL);
 #endif
 #ifndef	_NODOSDRIVE
-	if ((drive = tmpdosdupl("", &dir, 1)) < 0) return(FNC_CANCEL);
+	if ((drive = tmpdosdupl(nullstr, &dir, 1)) < 0) return(FNC_CANCEL);
 #endif
 	if (!execenv("FD_EDITOR", filelist[filepos].name)) {
 #ifdef	EDITOR
@@ -1202,7 +1215,7 @@ char *arg;
 	int fs;
 
 	if (writefs >= 2 || findpattern) return(warning_bell(arg));
-	if ((fs = writablefs(".")) <= 0) {
+	if ((fs = writablefs(curpath)) <= 0) {
 		warning(0, NOWRT_K);
 		return(FNC_CANCEL);
 	}
@@ -1225,7 +1238,7 @@ char *arg;
 
 	checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
 #ifndef	_NODOSDRIVE
-	if ((drive = dospath3(""))) flushdrv(drive, NULL);
+	if ((drive = dospath3(nullstr))) flushdrv(drive, NULL);
 #endif
 
 	return(FNC_EFFECT);
@@ -1253,7 +1266,7 @@ char *arg;
 #ifdef	_NOARCHIVE
 	help(0);
 #else
-	help(archivefile != NULL);
+	help((archivefile) ? 1 : 0);
 #endif
 
 	return(FNC_UPDATE);
@@ -1264,6 +1277,7 @@ static int confirmpty(VOID_A)
 {
 	int i;
 
+	VOID_C checkallpty();
 	for (i = 0; i < MAXWINDOWS; i++) if (ptylist[i].pid) break;
 	if (i < MAXWINDOWS && !yesno(KILL_K)) return(1);
 
@@ -1395,7 +1409,7 @@ char *arg;
 	}
 	if (strdelim(file, 0)) {
 		free(file);
-		file = strdup2("..");
+		file = strdup2(parentpath);
 	}
 	replacefname(file);
 
@@ -1408,17 +1422,18 @@ char *arg;
 	char *cp;
 	int len;
 
-	removepolicy = 0;
 	if (mark > 0) {
 		if (!yesno(DELMK_K)) return(FNC_CANCEL);
+		if (prepareremove(0, mark) < 0) return(FNC_CANCEL);
 		filepos = applyfile(rmvfile, NULL);
 	}
 	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	else {
 		cp = DELFL_K;
-		len = strlen2(cp) - ((int)sizeof("%.*s") - 1);
+		len = strlen2(cp) - strsize("%.*s");
 		if (!yesno(cp, n_lastcolumn - len, filelist[filepos].name))
 			return(FNC_CANCEL);
+		if (prepareremove(0, 1) < 0) return(FNC_CANCEL);
 		filepos = applyfile(rmvfile, NULL);
 	}
 	if (filepos >= maxfile && (filepos -= 2) < 0) filepos = 0;
@@ -1429,32 +1444,32 @@ char *arg;
 static int delete_dir(arg)
 char *arg;
 {
+#if	!defined (NOSYMLINK) && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	char *cp;
-	int len;
+	int ret, len;
 
 	if (!isdir(&(filelist[filepos])) || isdotdir(filelist[filepos].name))
 		return(warning_bell(arg));
 	cp = DELDR_K;
-	len = strlen2(cp) - ((int)sizeof("%.*s") - 1);
+	len = strlen2(cp) - strsize("%.*s");
 	if (!yesno(cp, n_lastcolumn - len, filelist[filepos].name))
 		return(FNC_CANCEL);
-	removepolicy = 0;
 	cp = filelist[filepos].name;
 #ifndef	NOSYMLINK
 	if (islink(&(filelist[filepos]))) {
-# ifndef	_NODOSDRIVE
-		char path[MAXPATHLEN];
-# endif
-		int ret;
-
+		if (prepareremove(0, 1) < 0) return(FNC_CANCEL);
 		ret = rmvfile(fnodospath(path, filepos));
-		if (ret < 0) warning(-1, cp);
-		else if (!ret) filepos++;
+		if (ret == APL_ERROR) warning(-1, cp);
 	}
 	else
 #endif
-	if (!applydir(cp, rmvfile, NULL, rmvdir, ORD_NOPREDIR, NULL))
-		filepos++;
+	{
+		if (prepareremove(1, 1) < 0) return(FNC_CANCEL);
+		ret = applydir(cp, rmvfile, NULL, rmvdir, ORD_NOPREDIR, NULL);
+	}
+	if (ret == APL_OK) filepos++;
 	if (filepos >= maxfile && (filepos -= 2) < 0) filepos = 0;
 
 	return(FNC_EFFECT);
@@ -1495,8 +1510,8 @@ char *arg;
 
 	if (!(findregexp = prepareregexp(FINDD_K, arg))) return(FNC_CANCEL);
 	destpath = NULL;
-	cp = isdir(&(filelist[filepos])) ? filelist[filepos].name : ".";
-	applydir(cp, findfile, finddir, NULL, ORD_NORMAL, NOFND_K);
+	cp = isdir(&(filelist[filepos])) ? filelist[filepos].name : curpath;
+	VOID_C applydir(cp, findfile, finddir, NULL, ORD_NORMAL, NOFND_K);
 	regexp_free(findregexp);
 	if (!destpath) return(FNC_CANCEL);
 
@@ -1516,7 +1531,8 @@ int evalstatus(n)
 int n;
 {
 	if (n < 0) return(FNC_CANCEL);
-	if (internal_status > FNC_FAIL) return(internal_status);
+	if (internal_status == FNC_QUIT || internal_status > FNC_CANCEL)
+		return(internal_status);
 
 	return(FNC_EFFECT);
 }
@@ -1615,7 +1631,7 @@ char *arg;
 	else
 #endif
 #ifndef	_NODOSDRIVE
-	if ((drive = tmpdosdupl("", &dir, 1)) < 0) ret = -1;
+	if ((drive = tmpdosdupl(nullstr, &dir, 1)) < 0) ret = -1;
 	else if (drive) {
 		ret = ptyusercomm(com, filelist[filepos].name,
 			F_ARGSET | F_IGNORELIST);
@@ -1720,7 +1736,7 @@ char *arg;
 		return(FNC_CANCEL);
 	else if (!*(path = evalpath(path, 0))) {
 		free(path);
-		path = strdup2(".");
+		path = strdup2(curpath);
 	}
 
 	ret = infofs(path);
@@ -1740,11 +1756,21 @@ char *s;
 	str[1] = CDATE_K;
 	val[0] = ATR_MODEONLY;
 	val[1] = ATR_TIMEONLY;
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+	str[2] = COWNR_K;
+	val[2] = ATR_OWNERONLY;
+#endif
 
 	n = ATR_MODEONLY;
 	Xlocate(0, ATTR_Y);
 	Xputterm(L_CLEAR);
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+	Xputterm(T_STANDOUT);
+#endif
 	Xkanjiputs(s);
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+	Xputterm(END_STANDOUT);
+#endif
 	if (selectstr(&n, MAXATTRSEL, ATTR_X, str, val) != K_CR) return(-1);
 
 	return(n);
@@ -1784,6 +1810,30 @@ char *arg;
 	return(FNC_EFFECT);
 }
 
+#ifndef	_NOEXTRAATTR
+/*ARGSUSED*/
+static int attr_dir(arg)
+char *arg;
+{
+	int n, flag;
+
+	if (!isdir(&(filelist[filepos]))) return(warning_bell(arg));
+
+	if ((flag = selectattr(ATTRD_K)) < 0) return(FNC_CANCEL);
+	flag |= (ATR_MULTIPLE | ATR_RECURSIVE);
+	while ((n = inputattr(&(filelist[filepos]), flag)) < 0)
+		warning(0, ILTMS_K);
+	if (!n) {
+		if (FILEPERROW < WFILEMINATTR) return(FNC_EFFECT);
+		return(FNC_UPDATE);
+	}
+	VOID_C applydir(filelist[filepos].name, setattr,
+		NULL, setattr, ORD_NOPREDIR, NULL);
+
+	return(FNC_EFFECT);
+}
+#endif	/* !_NOEXTRAATTR */
+
 #ifndef	_NOTREE
 /*ARGSUSED*/
 static int tree_dir(arg)
@@ -1798,8 +1848,10 @@ char *arg;
 		return(FNC_UPDATE);
 	}
 	free(path);
-	if (findpattern) free(findpattern);
-	findpattern = NULL;
+	if (findpattern) {
+		free(findpattern);
+		findpattern = NULL;
+	}
 	replacefname(NULL);
 
 	return(FNC_EFFECT);
@@ -1843,7 +1895,7 @@ char *arg;
 		}
 	}
 #ifndef	_NODOSDRIVE
-	else if ((drive = tmpdosdupl("", &dir, 0)) < 0) {
+	else if ((drive = tmpdosdupl(nullstr, &dir, 0)) < 0) {
 		free(dev);
 		return(FNC_CANCEL);
 	}
@@ -1904,9 +1956,9 @@ int oldwin;
 			&= ~(F_ISMRK | F_WSMRK | F_ISARG);
 	if (filelist) return;
 	maxfile = winvar[oldwin].v_maxfile;
+	i = ((maxfile) ? maxfile : 1) * sizeof(namelist);
 	addlist();
-	memcpy((char *)filelist, (char *)(winvar[oldwin].v_filelist),
-		winvar[oldwin].v_maxfile * sizeof(namelist));
+	memcpy((char *)filelist, (char *)(winvar[oldwin].v_filelist), i);
 	for (i = 0; i < winvar[oldwin].v_maxfile; i++)
 		filelist[i].name = strdup2(winvar[oldwin].v_filelist[i].name);
 	filepos = winvar[oldwin].v_filepos;
@@ -2015,7 +2067,7 @@ char *arg;
 {
 	if (nextwin() < 0) return(FNC_NONE);
 # ifndef	_NOPTY
-	if (ptylist[win].pid && ptylist[win].status < 0) {
+	if (!ptyinternal && ptylist[win].pid && ptylist[win].status < 0) {
 		rewritefile(0);
 		VOID_C frontend();
 		if (internal_status > FNC_FAIL) return(internal_status);
@@ -2096,7 +2148,7 @@ char *arg;
 	if (chdir3(fullpath, 1) < 0) lostcwd(fullpath);
 
 # ifndef	_NOPTY
-	if (ptylist[win].pid && ptylist[win].status < 0) {
+	if (!ptyinternal && ptylist[win].pid && ptylist[win].status < 0) {
 		rewritefile(0);
 		VOID_C frontend();
 		if (internal_status > FNC_FAIL) return(internal_status);
