@@ -8,13 +8,10 @@
 #ifdef	FD
 #include "fd.h"
 #else	/* !FD */
+#include "machine.h"
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include "machine.h"
-#include "printf.h"
-#include "kctype.h"
-#include "pathname.h"
 
 #ifndef	NOUNISTDH
 #include <unistd.h>
@@ -23,6 +20,10 @@
 #ifndef	NOSTDLIBH
 #include <stdlib.h>
 #endif
+
+#include "printf.h"
+#include "kctype.h"
+#include "pathname.h"
 #endif	/* !FD */
 
 #ifdef	USETIMEH
@@ -31,6 +32,11 @@
 
 #if	MSDOS && !defined (FD)
 #define	VOL_FAT32	"FAT32"
+# ifdef	DOUBLESLASH
+# define	MNTDIRSIZ	MAXPATHLEN
+# else
+# define	MNTDIRSIZ	(3 + 1)
+# endif
 #endif	/* MSDOS && !FD */
 
 #if	MSDOS
@@ -119,7 +125,7 @@ typedef struct fs_data		statfs_t;
 #    if	(STATFSARGS == 3)
 #    define	statfs2(path, buf)	statfs(path, buf, sizeof(statfs_t))
 #    else
-#    define	statfs2		statfs
+#    define	statfs2			statfs
 #    endif
 #   endif
 #  endif	/* USESTATFSH || USEVFSH || USEMOUNTH */
@@ -293,19 +299,20 @@ struct filestat_t {
 # else
 # define	L_SET	0
 # endif
-#endif
+#endif	/* !L_SET */
 #ifndef	L_INCR
 # ifdef	SEEK_CUR
 # define	L_INCR	SEEK_CUR
 # else
 # define	L_INCR	1
 # endif
-#endif
+#endif	/* !L_INCR */
 
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strdup2 __P_((char *));
+extern char *strncpy2 __P_((char *, char *, int));
 
 #ifdef	FD
 extern int getinfofs __P_((char *, off_t *, off_t *, off_t *));
@@ -330,7 +337,7 @@ static int NEAR Xutimes __P_((char *, struct timeval []));
 #  endif
 # define	ttyiomode(n)
 # define	stdiomode()
-# define	getkey2(n)	getch();
+# define	getkey3(n, c)	getch();
 # else	/* !MSDOS */
 # define	realpath2(p, r, f) \
 				realpath(p, r)
@@ -339,7 +346,7 @@ static int NEAR Xutimes __P_((char *, struct timeval []));
 static VOID NEAR ttymode __P_((int));
 # define	ttyiomode(n)	(ttymode(1))
 # define	stdiomode()	(ttymode(0))
-static int NEAR getkey2 __P_((int));
+static int NEAR getkey3 __P_((int, int));
 # endif	/* !MSDOS */
 static int NEAR touchfile __P_((char *, struct stat *));
 #ifndef	NODIRLOOP
@@ -411,7 +418,7 @@ extern char *sys_errlist[];
 #endif
 
 static CONST char *doserrstr[] = {
-	"",
+	NULL,
 #define	ER_REQPARAM	1
 	"Required parameter missing",
 #define	ER_TOOMANYPARAM	2
@@ -427,10 +434,10 @@ static CONST char *doserrstr[] = {
 #define	ER_INVALIDSW	7
 	"Invalid switch",
 };
-#define	DOSERRSIZ	((int)(sizeof(doserrstr) / sizeof(char *)))
+#define	DOSERRSIZ	arraysize(doserrstr)
 
-static char dirsort[(sizeof(DIRSORTFLAG) - 1) * 2 + 1] = "";
-static char dirattr[(sizeof(DIRATTRFLAG) - 1) * 2 + 1] = "";
+static char dirsort[strsize(DIRSORTFLAG) * 2 + 1] = "";
+static char dirattr[strsize(DIRATTRFLAG) * 2 + 1] = "";
 static int dirline = 0;
 static char *dirwd = NULL;
 static int dirtype = '\0';
@@ -454,17 +461,30 @@ char *path;
 off_t *totalp, *freep, *bsizep;
 {
 # if	MSDOS
+#  ifdef	DOUBLESLASH
+	char *cp;
+	int len;
+#  endif
 	struct SREGS sreg;
 	__dpmi_regs reg;
-	char drv[4], buf[128];
-# endif
+	char drv[MNTDIRSIZ], buf[128];
+# endif	/* !MSDOS */
 	statfs_t fsbuf;
 
 # if	MSDOS
-	drv[0] = toupper2(path[0]);
-	drv[1] = ':';
-	drv[2] = '\\';
-	drv[3] = '\0';
+#  ifdef	DOUBLESLASH
+	if ((len = isdslash(path))) {
+		if ((cp = strdelim(&(path[len]), 0))) {
+			len = cp - path;
+			if (!(cp = strdelim(&(path[len + 1]), 0)))
+				cp += strlen(cp);
+			len = cp - path;
+		}
+		strncpy2(drv, path, len);
+	}
+	else
+#  endif
+	VOID_C gendospath(drv, toupper2(path[0]), _SC_);
 
 	reg.x.ax = 0x71a0;
 	reg.x.bx = 0;
@@ -636,8 +656,8 @@ int on;
 }
 
 /*ARGSUSED*/
-static int NEAR getkey2(sig)
-int sig;
+static int NEAR getkey3(sig, code)
+int sig, code;
 {
 	u_char ch;
 	int i;
@@ -692,7 +712,7 @@ char *src, *dest;
 	char path[MAXPATHLEN];
 	int len;
 
-	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if ((len = Xreadlink(src, path, strsize(path))) < 0) return(-1);
 	if (Xlstat(dest, &st) >= 0) {
 #  ifndef	NODIRLOOP
 		if (issamebody(src, dest)) return(0);
@@ -757,7 +777,7 @@ int n;
 		kanjifputs(s, stderr);
 		fputnl(stderr);
 	}
-	fputs(doserrstr[n], stderr);
+	if (doserrstr[n]) fputs(doserrstr[n], stderr);
 	fputnl(stderr);
 }
 
@@ -799,7 +819,7 @@ static int NEAR inputkey(VOID_A)
 	int c;
 
 	ttyiomode(1);
-	c = getkey2(0);
+	c = getkey3(0, inputkcode);
 	stdiomode();
 	if (c == EOF) c = -1;
 	else if (c == cc_intr) {
@@ -848,7 +868,9 @@ CONST VOID_P vp2;
 				else ret = (cp2) ? -1 : 0;
 				break;
 			case 'D':
-				ret = sp1 -> mtim - sp2 -> mtim;
+				if (sp1 -> mtim == sp2 -> mtim) ret = 0;
+				else ret = (sp1 -> mtim > sp2 -> mtim)
+					? 1 : -1;
 				break;
 			case 'G':
 				ret = dir_isdir(sp2) - dir_isdir(sp1);
@@ -1199,15 +1221,17 @@ struct filestat_t *dirp;
 static VOID NEAR showfnameb(dirp)
 struct filestat_t *dirp;
 {
-	char buf[MAXPATHLEN];
+	char *cp, buf[MAXPATHLEN];
 	int i;
 
 	if (dirflag & DF_SUBDIR) {
+		cp = dirwd;
 		if (dirflag & DF_LOWER) {
 			for (i = 0; dirwd[i]; i++) buf[i] = tolower2(dirwd[i]);
-			dirwd = buf;
+			buf[i] = '\0';
+			cp = buf;
 		}
-		fprintf2(stdout, "%k%c", dirwd, _SC_);
+		fprintf2(stdout, "%k%c", cp, _SC_);
 	}
 
 	if (dirflag & DF_LOWER)
@@ -1536,7 +1560,7 @@ char *argv[];
 	if ((n = getdiropt(argc, argv)) < 0)
 		return(RET_FAIL);
 	if (argc <= n) {
-		dir = ".";
+		dir = curpath;
 		file = "*";
 	}
 	else if (argc > n + 1) {
@@ -1552,12 +1576,12 @@ char *argv[];
 	else {
 		strcpy(buf, argv[n]);
 		if (!(file = strrdelim(buf, 1))) {
-			dir = ".";
+			dir = curpath;
 			file = argv[n];
 		}
 		else {
 			i = file - buf;
-			if (isrootdir(buf) || *file != _SC_) file++;
+			if (isrootpath(buf) || *file != _SC_) file++;
 			*file = '\0';
 			dir = buf;
 			file = &(argv[n][++i]);
@@ -1570,11 +1594,7 @@ char *argv[];
 		return(RET_FAIL);
 	}
 	if (dir != buf) strcpy(wd, cwd);
-	else {
-		if (chdir2(buf) < 0) {
-			dosperror(buf);
-			return(RET_FAIL);
-		}
+	else if (chdir2(buf) >= 0) {
 		if (!Xgetwd(wd)) {
 			dosperror(NULL);
 			return(RET_FAIL);
@@ -1584,6 +1604,13 @@ char *argv[];
 			return(RET_FAIL);
 		}
 	}
+#ifdef	DOUBLESLASH
+	else if (isdslash(buf)) realpath2(buf, wd, 1);
+#endif
+	else {
+		dosperror(buf);
+		return(RET_FAIL);
+	}
 	if (getinfofs(wd, &total, &fre, &bsize) < 0) total = fre = (off_t)-1;
 
 	if (dirtype != 'B') fputnl(stdout);
@@ -1666,7 +1693,7 @@ int argc;
 char *argv[];
 {
 	char *buf, **wild;
-	int i, n, key, flag, ret;
+	int i, n, c, flag, ret;
 
 	flag = 0;
 	for (n = 1; n < argc; n++) {
@@ -1704,10 +1731,10 @@ char *argv[];
 			stdiomode();
 			if (!buf) return(RET_SUCCESS);
 			if (!isatty(STDIN_FILENO)) fputnl(stdout);
-			key = *buf;
+			c = *buf;
 			free(buf);
-		} while (!strchr("ynYN", key));
-		if (key == 'n' || key == 'N') return(RET_SUCCESS);
+		} while (!strchr("ynYN", c));
+		if (c == 'n' || c == 'N') return(RET_SUCCESS);
 	}
 	if (!(wild = evalwild(argv[n], 0))) {
 		doserror(argv[n], ER_FILENOTFOUND);
@@ -1720,16 +1747,15 @@ char *argv[];
 				fprintf2(stdout,
 					"%k,    Delete (Y/N)?", wild[i]);
 				fflush(stdout);
-				if ((key = inputkey()) < 0) {
+				if ((c = inputkey()) < 0) {
 					freevar(wild);
 					return(ret);
 				}
-				if (key <= (int)MAXUTYPE(u_char)
-				&& isprint2(key))
-					fputc(key, stdout);
+				if (c <= (int)MAXUTYPE(u_char) && isprint2(c))
+					fputc(c, stdout);
 				fputnl(stdout);
-			} while (!strchr("ynYN", key));
-			if (key == 'n' || key == 'N') continue;
+			} while (!strchr("ynYN", c));
+			if (c == 'n' || c == 'N') continue;
 		}
 		if (Xunlink(wild[i]) < 0) {
 			dosperror(wild[i]);
@@ -1872,15 +1898,9 @@ int bin;
 	char *cp;
 
 	if (bin < 0) return(-1);
-	if ((cp = strchr(name, DOSCOMOPT))) {
-		if (!cp[1] || (cp[2] && cp[3])) {
-			doserror(cp, ER_INVALIDSW);
-			return(-1);
-		}
-		*(cp++) = '\0';
-		if (toupper2(*cp) == 'B') bin = CF_BINARY;
-		else if (toupper2(*cp) == 'A') bin = CF_TEXT;
-	}
+	if (!(cp = strchr(name, DOSCOMOPT)) || !cp[1] || cp[2]) /*EMPTY*/;
+	else if (toupper2(*cp) == 'B') bin = CF_BINARY;
+	else if (toupper2(*cp) == 'A') bin = CF_TEXT;
 
 	return(bin);
 }
@@ -2081,9 +2101,11 @@ int sbin, dbin, dfd;
 	}
 
 #ifdef	FD
-	stp -> st_nlink = (TCH_ATIME | TCH_MTIME);
-#endif
+	stp -> st_nlink = (TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	if (!tty && touchfile(dest, stp) < 0) return(-1);
+#else
+	if (!tty) VOID_C touchfile(dest, stp);
+#endif
 
 	return(1);
 }
