@@ -13,6 +13,10 @@
 #include "kanji.h"
 #include "termemu.h"
 
+#ifndef	_NOIME
+#include "roman.h"
+#endif
+
 #ifdef	_NOORIGSHELL
 #include "termio.h"
 #include "wait.h"
@@ -24,20 +28,19 @@ extern int deletealias __P_((char *));
 # endif
 #endif
 
-#ifdef	SIGALRM
-#define	sigalrm()	((!noalrm) ? SIGALRM : 0)
+#ifdef	_NOIME
+#define	MAXPTYMENU	4
 #else
-#define	sigalrm()	0
+#define	MAXPTYMENU	5
 #endif
 
 extern functable funclist[];
 extern int internal_status;
-#ifdef	SIGALRM
-extern int noalrm;
-#endif
 extern int fdmode;
 extern int fdflags;
 extern int wheader;
+extern int lcmdline;
+extern int maxcmdline;
 extern int ptymode;
 extern int ptyinternal;
 extern int ptymenukey;
@@ -45,6 +48,9 @@ extern p_id_t emupid;
 extern int emufd;
 extern int parentfd;
 extern char *ptytmpfile;
+#ifndef	_NOIME
+extern int ime_cont;
+#endif
 
 static int (*lastfunc)__P_((VOID_A)) = NULL;
 static u_long lockflags = 0;
@@ -59,6 +65,7 @@ static int NEAR recvcommand __P_((int, command_t **, syntaxtree *));
 static int NEAR recvstree __P_((int, syntaxtree **, syntaxtree *));
 #endif
 static VOID NEAR recvchild __P_((int));
+static VOID NEAR sendkey __P_((int));
 
 
 int waitstatus(pid, options, statusp)
@@ -91,7 +98,7 @@ static int waitpty(VOID_A)
 			if (result[i] && ptylist[i].pid) recvchild(i);
 	}
 
-	if (win != oldwin) {
+	if (win != oldwin || !emupid) {
 		recvchild(oldwin);
 		return(-2);
 	}
@@ -245,12 +252,7 @@ va_dcl
 	va_end(args);
 	if (n < 0) error("malloc()");
 
-	if (!emupid) VOID_C cputs2(buf);
-	else {
-		sendword(emufd, TE_CPUTS2);
-		sendword(emufd, n);
-		sendbuf(emufd, buf, n);
-	}
+	Xcputs2(buf);
 	free(buf);
 
 	return(n);
@@ -373,7 +375,7 @@ VOID changekcode(VOID_A)
 
 VOID changeinkcode(VOID_A)
 {
-	ptylist[win].incode = ptyinkcode;
+	ptylist[win].incode = (u_char)ptyinkcode;
 	if (!emupid) return;
 
 	sendword(emufd, TE_CHANGEINKCODE);
@@ -383,7 +385,7 @@ VOID changeinkcode(VOID_A)
 
 VOID changeoutkcode(VOID_A)
 {
-	ptylist[win].outcode = ptyoutkcode;
+	ptylist[win].outcode = (u_char)ptyoutkcode;
 	if (!emupid) return;
 
 	sendword(emufd, TE_CHANGEOUTKCODE);
@@ -394,12 +396,37 @@ VOID changeoutkcode(VOID_A)
 
 static int NEAR ptygetkey(VOID_A)
 {
-	char *cp, *str[4];
-	int n, c, ch, val[4];
+#ifndef	_NOIME
+	static char buf[MAXKANJIBUF + 1] = "";
+	static int next = 0;
+	u_int w;
+#endif
+	char *cp, *str[MAXPTYMENU];
+	int n, c, ch, val[MAXPTYMENU];
+
+#ifndef	_NOIME
+	if (next > 0) {
+		if (next >= sizeof(buf) || !(c = (u_char)(buf[next++])))
+			next = 0;
+		else return(c);
+	}
+	if (ime_cont) {
+		c = ime_inputkanji(sigalrm(1), buf);
+		if (c < 0 || !ime_cont) movepos(filepos, 0);
+		if (c != K_ESC) {
+			if (c < 0) {
+				next = 0;
+				ime_inputkanji(0, NULL);
+			}
+			else if (!c && (c = (u_char)(buf[0]))) next = 1;
+			return(c);
+		}
+	}
+#endif	/* !_NOIME */
 
 	for (;;) {
 		n = -1;
-		c = getkey2(sigalrm());
+		c = getkey2(sigalrm(1), inputkcode);
 		while (lockflags & (1 << win)) {
 			kbhit2(1000000L / SENSEPERSEC);
 			waitpty();
@@ -407,50 +434,104 @@ static int NEAR ptygetkey(VOID_A)
 #ifndef	_NOORIGSHELL
 		if (isshptymode()) break;
 #endif
-		if (c < 0 || ptymenukey < 0 || c != ptymenukey) break;
+		if (c < 0 || ptymenukey < 0
+		|| (c & ~K_ALTERNATE) != ptymenukey)
+			break;
 
 		str[0] = asprintf3(PTYAI_K, getkeysym(ptymenukey, 0));
 		str[1] = PTYIC_K;
 		str[2] = PTYBR_K;
-		str[3] = PTYNW_K;
+#ifndef	_NOIME
+		str[3] = PTYKJ_K;
+#endif
+		str[MAXPTYMENU - 1] = PTYNW_K;
 		val[0] = 0;
 		val[1] = 1;
 		val[2] = 2;
+#ifndef	_NOIME
 		val[3] = 3;
+#endif
+		val[MAXPTYMENU - 1] = 4;
 
 		n = 0;
 		changewin(MAXWINDOWS, (p_id_t)-1);
-		ch = selectstr(&n, (windows > 1) ? 4 : 3, 0, str, val);
+		ch = selectstr(&n, (windows > 1) ? MAXPTYMENU : MAXPTYMENU - 1,
+			0, str, val);
+		movepos(filepos, 0);
 		changewin(win, (p_id_t)-1);
 		free(str[0]);
-		if (ch != K_CR) /*EMPTY*/;
+
+		if (ch != K_CR) continue;
 		else if (!n) break;
 		else if (n == 2) {
 			killpty(win, NULL);
+			rewritefile(1);
 			c = -1;
 			break;
 		}
+#ifndef	_NOIME
 		else if (n == 3) {
+			c = ime_inputkanji(sigalrm(1), buf);
+			if (c < 0 || !ime_cont) movepos(filepos, 0);
+			if (c != K_ESC) {
+				if (c < 0) {
+					next = 0;
+					ime_inputkanji(0, NULL);
+				}
+				else if (!c && (c = (u_char)(buf[0])))
+					next = 1;
+				break;
+			}
+		}
+#endif	/* !_NOIME */
+#ifndef	_NOSPLITWIN
+		else if (n == 4) {
 			VOID_C nextwin();
 			c = -2;
 			break;
 		}
+#endif
 		else {
 			changewin(MAXWINDOWS, (p_id_t)-1);
+			lcmdline = -1;
+			maxcmdline = 1;
 			cp = inputstr(PTYKC_K, 1, -1, NULL, -1);
+			movepos(filepos, 0);
 			changewin(win, (p_id_t)-1);
-			if (cp) {
-				c = getkeycode(cp, 0);
-				free(cp);
-				if (c >= 0) break;
-				warning(0, VALNG_K);
+			if (!cp) continue;
+#ifdef	_NOIME
+			c = getkeycode(cp, 0);
+			free(cp);
+			if (c >= 0) break;
+#else	/* !_NOIME */
+			w = ime_getkeycode(cp);
+			free(cp);
+			if (!w) /*EMPTY*/;
+			else if (!(w & ~01777)) {
+				c = w;
+				break;
 			}
-		}
+			else if ((c = ime_inkanjiconv(buf, w)) < 0) /*EMPTY*/;
+			else if (c || !(c = (u_char)(buf[0]))) break;
+			else {
+				next = 1;
+				break;
+			}
+#endif	/* !_NOIME */
 
-		rewritefile(1);
+			changewin(MAXWINDOWS, (p_id_t)-1);
+			warning(0, VALNG_K);
+			movepos(filepos, 0);
+			changewin(win, (p_id_t)-1);
+		}
 	}
 
-	if (n >= 0) rewritefile(1);
+#ifndef	_NOIME
+	if (c < 0) {
+		next = 0;
+		ime_inputkanji(0, NULL);
+	}
+#endif
 
 	return(c);
 }
@@ -625,7 +706,7 @@ int w;
 	fd = ptylist[w].pipe;
 	if (recvbuf(fd, &uc, sizeof(uc)) < 0) return;
 
-	switch(uc) {
+	switch (uc) {
 		case TE_SETVAR:
 			if (recvbuf(fd, &varp, sizeof(varp)) < 0
 			|| recvvar(fd, &var) < 0)
@@ -670,7 +751,6 @@ int w;
 			setenv2(cp, func1, n);
 			free(cp);
 			if (func1) free(func1);
-			evalenv();
 			break;
 		case TE_ADDFUNCTION:
 			if (recvvar(fd, &var) < 0) break;
@@ -691,21 +771,18 @@ int w;
 			|| recvstring(fd, &cp) < 0 || !cp)
 				break;
 			if (putexportvar(cp, n) < 0) free(cp);
-			evalenv();
 			break;
 		case TE_PUTSHELLVAR:
 			if (recvbuf(fd, &n, sizeof(n)) < 0
 			|| recvstring(fd, &cp) < 0 || !cp)
 				break;
 			if (putshellvar(cp, n) < 0) free(cp);
-			evalenv();
 			break;
 		case TE_UNSET:
 			if (recvbuf(fd, &n, sizeof(n)) < 0
 			|| recvstring(fd, &cp) < 0 || !cp)
 				break;
 			VOID_C unset(cp, n);
-			evalenv();
 			free(cp);
 			break;
 		case TE_SETEXPORT:
@@ -761,7 +838,7 @@ int w;
 			|| recvbuf(fd, &val, sizeof(val)) < 0
 			|| recvstring(fd, &cp) < 0 || !cp)
 				break;
-			entryhist(n, cp, val);
+			VOID_C entryhist(n, cp, val);
 			free(cp);
 			break;
 		case TE_ADDKEYBIND:
@@ -789,7 +866,7 @@ int w;
 			|| recvbuf(fd, &(key.len), sizeof(key.len)) < 0
 			|| recvstring(fd, &(key.str)) < 0)
 				break;
-			setkeyseq(key.code, key.str, key.len);
+			VOID_C setkeyseq(key.code, key.str, key.len);
 			break;
 #ifndef	_NOARCHIVE
 		case TE_ADDLAUNCH:
@@ -834,11 +911,11 @@ int w;
 			|| recvbuf(fd, &dev, sizeof(dev)) < 0
 			|| recvstring(fd, &(dev.name)) < 0)
 				break;
-			insertdrv(n, &dev);
+			VOID_C insertdrv(n, &dev);
 			break;
 		case TE_DELETEDRV:
 			if (recvbuf(fd, &n, sizeof(n)) < 0) break;
-			deletedrv(n);
+			VOID_C deletedrv(n);
 			break;
 #endif	/* _USEDOSEMU */
 		case TE_LOCKFRONT:
@@ -871,6 +948,20 @@ int w;
 			}
 			else if (cp) free(cp);
 			break;
+#ifndef	_NOIME
+		case TE_ADDROMAN:
+			if (recvstring(fd, &cp) < 0 || !cp) break;
+			if (recvstring(fd, &func1) < 0) {
+				free(cp);
+				break;
+			}
+			VOID_C addroman(cp, func1);
+			break;
+		case TE_FREEROMAN:
+			if (recvbuf(fd, &n, sizeof(n)) < 0) break;
+			freeroman(n);
+			break;
+#endif	/* !_NOIME */
 		case TE_INTERNAL:
 			if (recvbuf(fd, &n, sizeof(n)) < 0
 			|| recvstring(fd, &cp) < 0)
@@ -900,6 +991,58 @@ int w;
 	}
 }
 
+static VOID NEAR sendkey(c)
+int c;
+{
+	static char buf[MAXUTF8LEN * sizeof(short)];
+	static int index = 0;
+	static int max = 0;
+#ifndef	_NOKANJICONV
+	int code;
+#endif
+	short w;
+
+#ifndef	_NOKANJICONV
+	code = (inputkcode != NOCNV) ? inputkcode : DEFCODE;
+#endif
+	if (c >= 0) {
+		w = c;
+		memcpy(&(buf[index++ * sizeof(w)]), (char *)&w, sizeof(w));
+	}
+
+	if (c < 0 || c & 0xff00) /*EMPTY*/;
+	else if (max && index >= max) /*EMPTY*/;
+#ifdef	_NOKANJICONV
+# ifdef	CODEEUC
+	else if (c == C_EKANA) {
+		max = 2;
+		return;
+	}
+# endif
+#else	/* !_NOKANJICONV */
+	else if (code == EUC && c == C_EKANA) {
+		max = 2;
+		return;
+	}
+	else if (code >= UTF8) {
+		if (index > 1) {
+			if ((c & 0xc0) == 0x80) return;
+		}
+		else if (ismsb(c)) {
+			max = ((c & 0xf0) == 0xe0) ? 3 : 2;
+			return;
+		}
+	}
+#endif	/* !_NOKANJICONV */
+	else if (isinkanji1(c, code)) {
+		max = 2;
+		return;
+	}
+
+	if (index) sendbuf(emufd, buf, index * sizeof(w));
+	index = max = 0;
+}
+
 int frontend(VOID_A)
 {
 	int n, ch, wastty, status;
@@ -913,7 +1056,8 @@ int frontend(VOID_A)
 
 	internal_status = FNC_FAIL;
 	if (!(wastty = isttyiomode)) Xttyiomode(1);
-	while ((ch = ptygetkey()) >= 0) sendword(emufd, ch);
+	while ((ch = ptygetkey()) >= 0) sendkey(ch);
+	sendkey(-1);
 	if (!wastty) Xstdiomode();
 
 	if (ch < -1) {
