@@ -103,9 +103,14 @@ typedef struct fstab		mnt_t;
 #endif	/* USEGETFSENT */
 
 #if	MSDOS
+# ifdef	DOUBLESLASH
+# define	MNTDIRSIZ	MAXPATHLEN
+# else
+# define	MNTDIRSIZ	(3 + 1)
+# endif
 typedef struct _mnt_t {
 	char *mnt_fsname;
-	char mnt_dir[4];
+	char mnt_dir[MNTDIRSIZ];
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
@@ -201,10 +206,12 @@ extern int unixstatfs __P_((char *, statfs_t *));
 extern VOID error __P_((char *));
 extern int _chdir2 __P_((char *));
 extern char *strcpy2 __P_((char *, char *));
+extern char *strncpy2 __P_((char *, char *, int));
 extern char *getwd2 __P_((VOID_A));
 extern VOID warning __P_((int, char *));
 #ifdef	_USEDOSPATH
 extern int dospath __P_((char *, char *));
+extern char *gendospath __P_((char *, int, int));
 #endif
 #if	MSDOS && !defined (_NOUSELFN)
 extern int supportLFN __P_((char *));
@@ -217,7 +224,7 @@ extern int dosstatfs __P_((int, char *));
 #endif
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
-#if	!MSDOS
+#if	!MSDOS || !defined (NOFLOCK)
 extern int Xstat __P_((char *, struct stat *));
 #endif
 extern int Xaccess __P_((char *, int));
@@ -318,6 +325,10 @@ extern int needbavail;
 #ifndef	MNTTYPE_FAT32
 #define	MNTTYPE_FAT32	"fat32"	/* Win98 */
 #endif
+#ifndef	MNTTYPE_SHARED
+#define	MNTTYPE_SHARED	"shared"	/* Win98 */
+#endif
+#define	MNTTYPE_XNFS	"nfs"	/* NFS */
 
 static int NEAR code2str __P_((char *, int));
 static int NEAR checkline __P_((int));
@@ -346,6 +357,7 @@ static CONST strtable mntlist[] = {
 	{FSID_FAT, MNTTYPE_PC},
 	{FSID_LFN, MNTTYPE_DOS7},
 	{FSID_LFN, MNTTYPE_FAT32},
+	{FSID_LFN, MNTTYPE_SHARED},
 #if	MSDOS
 # ifndef	_NODOSDRIVE
 	{FSID_FAT, MNTTYPE_FAT12},
@@ -438,8 +450,8 @@ int y;
 	return(y);
 }
 
-VOID help(mode)
-int mode;
+VOID help(arch)
+int arch;
 {
 	char buf[(KEYWID + 1 + 1) * 2 + 1];
 	int i, j, c, x, y, yy;
@@ -464,7 +476,7 @@ int mode;
 		Xlocate(x * (n_column / 2), yy + y);
 		if (x ^= 1) Xputterm(L_CLEAR);
 		else y++;
-		if (mode && !(funclist[i].status & FN_ARCHIVE)) continue;
+		if (arch && !(funclist[i].status & FN_ARCHIVE)) continue;
 
 		c = 0;
 		buf[0] = '\0';
@@ -755,25 +767,47 @@ char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
-	statfs_t fs;
-	mnt_t mnt;
 #if	MSDOS
 # ifndef	_NODOSDRIVE
 	int i;
 # endif
+# ifdef	DOUBLESLASH
+	char *cp;
+	int len;
+# endif
+	statfs_t fs;
+	mnt_t mnt;
+	char buf[MAXPATHLEN];
+	int drive;
 
 	if (!fsbuf) fsbuf = &fs;
 	if (!mntbuf) mntbuf = &mnt;
 
 	mntbuf -> mnt_fsname = "MSDOS";
-	mntbuf -> mnt_dir[0] = dospath(path, NULL);
-	mntbuf -> mnt_dir[1] = ':';
-	mntbuf -> mnt_dir[2] = _SC_;
-	mntbuf -> mnt_dir[3] = '\0';
+	drive = dospath(path, buf);
+# ifdef	DOUBLESLASH
+	if (drive == '_') {
+		if (*buf) path = buf;
+		len = isdslash(path);
+		if ((cp = strdelim(&(path[len]), 0))) {
+			len = cp - path;
+			if (!(cp = strdelim(&(path[len + 1]), 0)))
+				cp += strlen(cp);
+			len = cp - path;
+		}
+		strncpy2(mntbuf -> mnt_dir, path, len);
+	}
+	else
+# endif
+	VOID_C gendospath(mntbuf -> mnt_dir, drive, _SC_);
+
 # ifdef	_NOUSELFN
 	mntbuf -> mnt_type = MNTTYPE_PC;
 # else	/* !_NOUSELFN */
 	switch (supportLFN(mntbuf -> mnt_dir)) {
+		case 3:
+			mntbuf -> mnt_type = MNTTYPE_SHARED;
+			break;
 		case 2:
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
 			break;
@@ -810,13 +844,15 @@ mnt_t *mntbuf;
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) return(-1);
 #else	/* !MSDOS */
+# ifndef	_NODOSDRIVE
+	int drv;
+# endif
+	statfs_t fs;
+	mnt_t mnt;
 	mnt_t *mntp;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN];
 	ALLOC_T len, match;
-# ifndef	_NODOSDRIVE
-	int drv;
-# endif
 
 	if (!fsbuf) fsbuf = &fs;
 	if (!mntbuf) mntbuf = &mnt;
@@ -933,6 +969,28 @@ char *s1, *s2;
 	return(NULL);
 }
 
+#ifndef	NOFLOCK
+int isnfs(path)
+char *path;
+{
+	mnt_t mntbuf;
+	struct stat st;
+	char *cp, buf[MAXPATHLEN];
+
+	if (Xstat(path, &st) < 0) return(-1);
+	if (!s_isdir(&st)) {
+		if (!(cp = strrdelim(path, 1))) return(-1);
+		strncpy2(buf, path, cp - path);
+		path = buf;
+	}
+	if (getfsinfo(path, NULL, &mntbuf) < 0) return(-1);
+	if (strncasecmp2(mntbuf.mnt_type, MNTTYPE_XNFS, strsize(MNTTYPE_XNFS)))
+		return(0);
+
+	return(1);
+}
+#endif	/* NOFLOCK */
+
 int writablefs(path)
 char *path;
 {
@@ -1045,7 +1103,10 @@ off_t *totalp, *freep, *bsizep;
 #ifndef	_NODOSDRIVE
 	needbavail--;
 #endif
-	if (n < 0) return(-1);
+	if (n < 0) {
+		*bsizep = getblocksize(path);
+		return(-1);
+	}
 
 	*totalp = fsbuf.f_blocks;
 	*freep = fsbuf.f_bavail;
