@@ -243,6 +243,7 @@ K_EXTERN CONST u_char kctypetable[256]
 			(((u_char)(c1) & 0xf0) == 0xe0 \
 			&& ((u_char)(c2) & 0xc0) == 0x80 \
 			&& ((u_char)(c3) & 0xc0) == 0x80)
+#define	iswucs2(u)	(((u) & 0xff00) && ((u) < 0xff61 || (u) > 0xff9f))
 
 #define	NOCNV	0
 #define	ENG	1
@@ -258,6 +259,7 @@ K_EXTERN CONST u_char kctypetable[256]
 #define	CAP	11
 #define	UTF8	12
 #define	M_UTF8	13
+#define	I_UTF8	14
 
 #ifdef	CODEEUC
 #define	DEFCODE	EUC
@@ -269,6 +271,11 @@ K_EXTERN CONST u_char kctypetable[256]
 #define	KANAWID	1
 #endif
 
+#define	MAXKANJIBUF	(3 + 2 + 3)
+#define	MAXKLEN		2
+#define	MAXUTF8LEN	3
+#define	MAXNFLEN	4
+
 #ifdef	NOMULTIKANJI
 #define	_NOKANJICONV
 #define	_NOKANJIFCONV
@@ -284,38 +291,50 @@ K_EXTERN int kanjierrno K_INIT(0);
 #ifndef	_NOKANJIFCONV
 K_EXTERN int defaultkcode K_INIT(NOCNV);
 #endif
-#ifndef	_NOKANJICONV
+#ifdef	_NOKANJICONV
+#define	inputkcode	NOCNV
+#else
 K_EXTERN int inputkcode K_INIT(NOCNV);
 #endif
 #if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
 K_EXTERN int outputkcode K_INIT(NOCNV);
 #endif
+#if	!defined (_NOKANJICONV) && !defined (_NOPTY)
+K_EXTERN int ptyinkcode K_INIT(NOCNV);
+K_EXTERN int ptyoutkcode K_INIT(NOCNV);
+#endif
 #ifndef	_NOKANJIFCONV
 K_EXTERN int fnamekcode K_INIT(NOCNV);
 #endif
+#if	!defined (_NOENGMES) && !defined (_NOJPNMES)
+K_EXTERN int messagelang K_INIT(NOCNV);
+#endif
 
 #define	L_INPUT		0001
 #define	L_OUTPUT	0002
 #define	L_FNAME		0004
+#define	L_TERMINAL	0010
+#define	L_MESSAGE	0020
 
 K_EXTERN CONST char kanjiiomode[]
 # ifdef	K_INTERN
 = {
-	          L_OUTPUT | L_FNAME,	/* NOCNV */
-	          L_OUTPUT,		/* ENG */
-	L_INPUT | L_OUTPUT | L_FNAME,	/* SJIS */
-	L_INPUT | L_OUTPUT | L_FNAME,	/* EUC */
-	          L_OUTPUT | L_FNAME,	/* JIS7 */
-	          L_OUTPUT | L_FNAME,	/* O_JIS7 */
-	          L_OUTPUT | L_FNAME,	/* JIS8 */
-	          L_OUTPUT | L_FNAME,	/* O_JIS8 */
-	          L_OUTPUT | L_FNAME,	/* JUNET */
-	          L_OUTPUT | L_FNAME,	/* O_JUNET */
-	                     L_FNAME,	/* HEX */
-	                     L_FNAME,	/* CAP */
-	L_INPUT | L_OUTPUT | L_FNAME,	/* UTF8 */
-	L_INPUT | L_OUTPUT | L_FNAME,	/* M_UTF8 */
+	          L_OUTPUT | L_FNAME | L_TERMINAL | L_MESSAGE,	/* NOCNV */
+	          L_OUTPUT                        | L_MESSAGE,	/* ENG */
+	L_INPUT | L_OUTPUT | L_FNAME | L_TERMINAL,		/* SJIS */
+	L_INPUT | L_OUTPUT | L_FNAME | L_TERMINAL,		/* EUC */
+	          L_OUTPUT | L_FNAME,				/* JIS7 */
+	          L_OUTPUT | L_FNAME,				/* O_JIS7 */
+	          L_OUTPUT | L_FNAME,				/* JIS8 */
+	          L_OUTPUT | L_FNAME,				/* O_JIS8 */
+	          L_OUTPUT | L_FNAME,				/* JUNET */
+	          L_OUTPUT | L_FNAME,				/* O_JUNET */
+	                     L_FNAME,				/* HEX */
+	                     L_FNAME,				/* CAP */
+	L_INPUT | L_OUTPUT | L_FNAME | L_TERMINAL,		/* UTF8 */
+	L_INPUT | L_OUTPUT | L_FNAME | L_TERMINAL,		/* M_UTF8 */
+	L_INPUT | L_OUTPUT | L_FNAME | L_TERMINAL,		/* I_UTF8 */
 }
 # endif	/* K_INTERN */
 ;
@@ -334,19 +353,36 @@ int i;
 }
 #endif	/* K_INTERN */
 
+K_EXTERN int iskana1 __P_((char *, int *));
+#ifdef	K_INTERN
+int iskana1(s, ip)
+char *s;
+int *ip;
+{
+# ifdef	CODEEUC
+	int n;
+
+	if ((n = isekana(s, *ip))) (*ip)++;
+	return(n);
+# else
+	return(isskana(s, *ip));
+# endif
+}
+#endif	/* K_INTERN */
+
 #ifdef	_NOKANJICONV
 # ifdef	CODEEUC
-# define	isinkanji1(c)	iseuc(c)
-# define	isinkanji2(c)	iseuc(c)
+# define	isinkanji1(c,k)	iseuc(c)
+# define	isinkanji2(c,k)	iseuc(c)
 # else
-# define	isinkanji1(c)	issjis1(c)
-# define	isinkanji2(c)	issjis2(c)
+# define	isinkanji1(c,k)	issjis1(c)
+# define	isinkanji2(c,k)	issjis2(c)
 # endif
 #else	/* !_NOKANJICONV */
-#define	isinkanji1(c)	((inputkcode == EUC) ? iseuc(c) : \
-			((inputkcode == SJIS) ? issjis1(c) : 0))
-#define	isinkanji2(c)	((inputkcode == EUC) ? iseuc(c) : \
-			((inputkcode == SJIS) ? issjis2(c) : 0))
+#define	isinkanji1(c,k)	(((k) == EUC) ? iseuc(c) : \
+			(((k) == SJIS) ? issjis1(c) : 0))
+#define	isinkanji2(c,k)	(((k) == EUC) ? iseuc(c) : \
+			(((k) == SJIS) ? issjis2(c) : 0))
 #endif	/* !_NOKANJICONV */
 
 #endif	/* !__KCTYPE_H_ */
