@@ -160,16 +160,23 @@ int rdlink;
 	if (n == 1) {
 		cp = strrdelim(resolved, 0);
 		top = resolved;
-#if	MSDOS
-		top += 2;
-#else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
+#ifdef	_USEDOSEMU
 		if (_dospath(resolved)) top += 2;
 		else
-# endif
+#endif
+#ifdef	DOUBLESLASH
+		if ((n = isdslash(resolved)) && cp < &(resolved[n])) {
+			top = &(resolved[n - 1]);
+			if (*top != _SC_) return(resolved);
+		}
+		else
+#endif
+#if	MSDOS
+		top += 2;
+#else
 		if (rdlink && evallink(resolved, cp) > 0)
 			return(_realpath2(path, resolved, rdlink));
-#endif	/* !MSDOS */
+#endif
 		if (!cp || cp == top) {
 			*top = _SC_;
 			cp = top + 1;
@@ -195,42 +202,26 @@ char *realpath2(path, resolved, rdlink)
 char *path, *resolved;
 int rdlink;
 {
-#if	!MSDOS && !defined (_NODOSDRIVE)
+#ifdef	_USEDOSEMU
 	char *cp;
 	int duplastdrive;
 #endif
-#if	MSDOS || !defined (_NODOSDRIVE)
+#ifdef	_USEDOSPATH
 	int drv;
+#endif
+#ifdef	DOUBLESLASH
+	int n;
+#endif
+#if	MSDOS
+	int drive;
 #endif
 	char tmp[MAXPATHLEN];
 
 	strcpy(tmp, path);
 	path = tmp;
 
-#if	MSDOS
-	drv = dospath(nullstr, NULL);
-	if ((resolved[0] = _dospath(path))) path += 2;
-	if (*path == _SC_) {
-		if (!resolved[0]) resolved[0] = drv;
-		resolved[1] = ':';
-		resolved[2] = _SC_;
-		resolved[3] = '\0';
-	}
-	else if (resolved[0] && resolved[0] != drv) {
-		if (setcurdrv(resolved[0], 0) < 0) {
-			resolved[1] = ':';
-			resolved[2] = _SC_;
-			resolved[3] = '\0';
-		}
-		else {
-			if (!Xgetwd(resolved)) lostcwd(resolved);
-			if (setcurdrv(drv, 0) < 0) error("setcurdrv()");
-		}
-	}
-#else	/* !MSDOS */
-	if (*path == _SC_) copyrootpath(resolved);
-# ifndef	_NODOSDRIVE
-	else if ((drv = _dospath(path))) {
+#ifdef	_USEDOSEMU
+	if ((drv = _dospath(path))) {
 		path += 2;
 		if (*path == _SC_) cp = NULL;
 		else {
@@ -239,18 +230,42 @@ int rdlink;
 			cp = dosgetcwd(resolved, MAXPATHLEN - 1);
 			lastdrive = duplastdrive;
 		}
-		if (!cp) {
-			resolved[0] = drv;
-			resolved[1] = ':';
-			resolved[2] = _SC_;
-			resolved[3] = '\0';
-		}
+		if (!cp) VOID_C gendospath(resolved, drv, _SC_);
 	}
-# endif
+	else
+#endif
+#ifdef	DOUBLESLASH
+	if ((n = isdslash(path))) {
+		strncpy2(resolved, path, n);
+		path += n;
+		if (*path == _SC_) path++;
+	}
+	else
+#endif
+	{
+#if	MSDOS
+		drv = dospath(nullstr, NULL);
+		if ((drive = _dospath(path))) path += 2;
+		if (*path == _SC_) {
+			if (!drive) drive = drv;
+			VOID_C gendospath(resolved, drive, _SC_);
+		}
+		else if (drive && drive != drv) {
+			if (setcurdrv(drive, 0) < 0)
+				VOID_C gendospath(resolved, drive, _SC_);
+			else {
+				if (!Xgetwd(resolved)) lostcwd(resolved);
+				if (setcurdrv(drv, 0) < 0) error("setcurdrv()");
+			}
+		}
+#else	/* !MSDOS */
+		if (*path == _SC_) copyrootpath(resolved);
 #endif	/* !MSDOS */
-	else if (!rdlink && resolved != fullpath && *fullpath)
-		strcpy(resolved, fullpath);
-	else if (!Xgetwd(resolved)) copyrootpath(resolved);
+		else if (!rdlink && resolved != fullpath && *fullpath)
+			strcpy(resolved, fullpath);
+		else if (!Xgetwd(resolved)) copyrootpath(resolved);
+	}
+
 	norealpath++;
 	_realpath2(path, resolved, rdlink);
 	norealpath--;
@@ -273,7 +288,7 @@ char *path;
 	{
 		int fd, duperrno;
 
-		if ((fd = open(curpath, O_RDONLY, 0600)) < 0) {
+		if ((fd = open(curpath, O_RDONLY, 0666)) < 0) {
 # ifdef	CYGWIN
 			char upath[MAXPATHLEN], spath[MAXPATHLEN];
 			char tmp[MAXPATHLEN];
@@ -347,7 +362,7 @@ char *path;
 		unixpath = NULL;
 	}
 #endif
-	if (getconstvar("PWD")) setenv2("PWD", fullpath, 1);
+	if (getconstvar(ENVPWD)) setenv2(ENVPWD, fullpath, 1);
 #if	MSDOS
 	if (unixrealpath(fullpath, tmp)) strcpy(fullpath, tmp);
 #endif
@@ -369,7 +384,7 @@ int raw;
 	char *cwd;
 
 	cwd = path;
-	if (!raw && path[0] && !path[1]) switch(path[0]) {
+	if (!raw && path[0] && !path[1]) switch (path[0]) {
 		case '.':
 			cwd = NULL;
 			break;
@@ -643,11 +658,17 @@ int *lenp, ptr;
 	return(len);
 }
 
+#ifdef	CODEEUC
 int strlen2(s)
 char *s;
 {
-	return(snprintf2(NULL, 0, "%s", s));
+	int i, len;
+
+	for (i = len = 0; s[i]; i++, len++) if (isekana(s, i)) i++;
+
+	return(len);
 }
+#endif	/* CODEEUC */
 
 int strlen3(s)
 char *s;
@@ -819,6 +840,7 @@ int export;
 #ifdef	_NOORIGSHELL
 	if (export) environ = _putenv2(cp, environ);
 	else environ2 = _putenv2(cp, environ2);
+	evalenv(name, len);
 # ifndef	_NOPTY
 	sendparent(TE_PUTSHELLVAR, name, value, export);
 # endif
@@ -881,7 +903,7 @@ int flags;
 		fflush(stderr);
 		Xttyiomode(1);
 		keyflush();
-		getkey2(0);
+		getkey3(0, inputkcode);
 		Xstdiomode();
 		fputnl(stderr);
 	}
@@ -929,7 +951,7 @@ char *command;
 		fflush(stderr);
 		Xttyiomode(1);
 		keyflush();
-		getkey2(0);
+		getkey3(0, inputkcode);
 		Xstdiomode();
 		fputnl(stderr);
 		if (wasttyflags & F_TTYIOMODE)
