@@ -14,7 +14,7 @@
 # ifndef	DJGPP
 # include <dos.h>
 #  ifdef	__TURBOC__
-extern unsigned _stklen = 0x5800;
+extern u_int _stklen = 0x5800;
 #  define	harderr_t	void
 #  else
 #  define	harderr_t	int
@@ -32,6 +32,14 @@ extern unsigned _stklen = 0x5800;
 #define	isorgpid()	(mypid == orgpid)
 #endif
 
+#if	!defined (_NOIME) && defined (DEBUG)
+#include "roman.h"
+#endif
+
+#ifndef	DATADIR
+#define	DATADIR		progpath
+#endif
+
 #if	MSDOS
 # ifndef	BSPATHDELIM
 extern char *adjustpname __P_((char *));
@@ -58,8 +66,7 @@ extern int maxarchive;
 #endif
 extern int wheader;
 extern char fullpath[];
-extern char *histfile;
-extern int savehist;
+extern short savehist[];
 extern char *helpindex[];
 extern int subwindow;
 extern int win_x;
@@ -68,9 +75,12 @@ extern int win_y;
 extern int custno;
 #endif
 extern char *deftmpdir;
-#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
+#ifdef	_USEUNICODE
 extern char *unitblpath;
 #endif
+#ifndef	_NOIME
+extern char *dicttblpath;
+#endif
 #ifndef	_NOPTY
 extern int ptymode;
 extern int ptyinternal;
@@ -148,6 +158,9 @@ static VOID NEAR printtime __P_((int));
 #ifdef	SIGALRM
 static int trapalrm __P_((VOID_A));
 #endif
+#ifndef	_NODOSDRIVE
+static int wrap_intrkey __P_((VOID_A));
+#endif
 static char *NEAR getversion __P_((int *));
 #ifndef	_NOLOGGING
 static VOID NEAR startlog __P_((char **));
@@ -161,6 +174,7 @@ static int NEAR evaloption __P_((char *[]));
 static char *NEAR searchenv __P_((char *, char *[]));
 static VOID NEAR setexecname __P_((char *));
 static VOID NEAR setexecpath __P_((char *, char *[]));
+int main __P_((int, char *[], char *[]));
 
 char *origpath = NULL;
 char *progpath = NULL;
@@ -240,6 +254,7 @@ char *s;
 		fputs(s, stderr);
 		fputnl(stderr);
 	}
+	fclose(stderr);
 
 	if (isorgpid()) {
 		inittty(1);
@@ -496,13 +511,13 @@ VOID setlinecol(VOID_A)
 {
 	char buf[MAXLONGWIDTH + 1];
 
-	if (getconstvar("LINES")) {
+	if (getconstvar(ENVLINES)) {
 		snprintf2(buf, sizeof(buf), "%d", n_line);
-		setenv2("LINES", buf, 1);
+		setenv2(ENVLINES, buf, 1);
 	}
-	if (getconstvar("COLUMNS")) {
+	if (getconstvar(ENVCOLUMNS)) {
 		snprintf2(buf, sizeof(buf), "%d", n_column);
-		setenv2("COLUMNS", buf, 1);
+		setenv2(ENVCOLUMNS, buf, 1);
 	}
 }
 
@@ -593,7 +608,7 @@ int xmax, ymax;
 		Xputterm(T_BELL);
 		Xcputnl();
 		Xtflush();
-		if (kbhit2(1000000L) && getkey2(0) == K_ESC) {
+		if (kbhit2(1000000L) && getkey3(0, inputkcode) == K_ESC) {
 			errno = 0;
 			error(INTR_K);
 		}
@@ -725,6 +740,13 @@ static int trapalrm(VOID_A)
 }
 #endif	/* SIGALRM */
 
+#ifndef	_NODOSDRIVE
+static int wrap_intrkey(VOID_A)
+{
+	return(intrkey(-1));
+}
+#endif	/* !_NODOSDRIVE */
+
 int sigvecset(set)
 int set;
 {
@@ -746,7 +768,7 @@ int set;
 #endif
 #ifndef	_NODOSDRIVE
 		doswaitfunc = waitmes;
-		dosintrfunc = intrkey;
+		dosintrfunc = wrap_intrkey;
 #endif
 		status = 1;
 	}
@@ -812,7 +834,7 @@ VOID title(VOID_A)
 		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
+	cp = (iswellomit()) ? nullstr : " (c)1995-2007 T.Shirai  ";
 	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
 	while (i-- > 0) Xputch2(' ');
@@ -824,7 +846,7 @@ VOID title(VOID_A)
 #ifndef	_NOCUSTOMIZE
 VOID saveorigenviron(VOID_A)
 {
-	orighelpindex = copystrarray(NULL, helpindex, NULL, 10);
+	orighelpindex = copystrarray(NULL, helpindex, NULL, MAXHELPINDEX);
 	origbindlist = copybind(NULL, bindlist);
 # ifndef	_NOKEYMAP
 	origkeymaplist = copykeyseq(NULL);
@@ -936,7 +958,7 @@ int exist;
 #ifdef	_NOORIGSHELL
 # if	!MSDOS
 	tmp = NULL;
-	if (!exist && (tmp = getconstvar("TERM"))) {
+	if (!exist && (tmp = getconstvar(ENVTERM))) {
 		struct stat st;
 
 		cp = malloc2(strlen(file) + strlen(tmp) + 1 + 1);
@@ -972,7 +994,7 @@ int exist;
 		}
 
 		cp = line + strlen(line);
-		for (cp--; cp >= line && (*cp == ' ' || *cp == '\t'); cp--);
+		for (cp--; cp >= line && isblank2(*cp); cp--);
 		cp[1] = '\0';
 
 		cont = 0;
@@ -1067,7 +1089,6 @@ char *argv[];
 		setenv2(tmp, cp, 0);
 		free(tmp);
 	}
-	evalenv();
 
 	return(i);
 }
@@ -1108,7 +1129,7 @@ char *argv, *envp[];
 
 	if (!Xgetwd(buf)) error(NOCWD_K);
 	origpath = strdup2(buf);
-	if ((cp = searchenv("PWD", envp))) {
+	if ((cp = searchenv(ENVPWD, envp))) {
 		*fullpath = '\0';
 		realpath2(cp, fullpath, 0);
 		realpath2(fullpath, buf, 1);
@@ -1123,7 +1144,8 @@ char *argv, *envp[];
 	cp = argv;
 #else
 	if (strdelim(argv, 0)) cp = argv;
-	else if ((cp = searchenv("PATH", envp))) cp = searchexecpath(argv, cp);
+	else if ((cp = searchenv(ENVPATH, envp)))
+		cp = searchexecpath(argv, cp);
 	if (!cp) progpath = strdup2(origpath);
 	else
 #endif
@@ -1139,7 +1161,13 @@ char *argv, *envp[];
 VOID initfd(argv)
 char **argv;
 {
-	if (interactive) {
+	int i;
+
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	{
 #if	!MSDOS
 		if (adjtty) {
 			Xstdiomode();
@@ -1147,7 +1175,7 @@ char **argv;
 			Xttyiomode(0);
 		}
 #endif	/* !MSDOS */
-		loadhistory(0, histfile);
+		for (i = 0; i < 2; i++) loadhistory(i);
 		entryhist(1, origpath, 1);
 	}
 #ifndef	_NOLOGGING
@@ -1160,11 +1188,16 @@ VOID prepareexitfd(status)
 int status;
 {
 	char cwd[MAXPATHLEN];
+	int i;
 
 #ifndef	_NOLOGGING
 	endlog(status);
 #endif
-	if (interactive && savehist > 0) savehistory(0, histfile);
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	for (i = 0; i < 2; i++) savehistory(i);
 
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
@@ -1174,9 +1207,12 @@ int status;
 	}
 	free(origpath);
 	free(progname);
-#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
+#ifdef	_USEUNICODE
 	free(unitblpath);
 #endif
+#ifndef	_NOIME
+	free(dicttblpath);
+#endif
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
@@ -1190,7 +1226,7 @@ int status;
 # endif
 # ifndef	_NOCUSTOMIZE
 	if (orighelpindex) {
-		freestrarray(orighelpindex, 10);
+		freestrarray(orighelpindex, MAXHELPINDEX);
 		free(orighelpindex);
 	}
 	free(origbindlist);
@@ -1228,9 +1264,14 @@ int status;
 # ifndef	_NOROCKRIDGE
 	detranspath(NULL, NULL);
 # endif
-# if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
+# ifdef	_USEUNICODE
 	discardunitable();
 # endif
+# ifndef	_NOIME
+	ime_freebuf();
+	freeroman(0);
+	discarddicttable();
+# endif
 #endif	/* DEBUG */
 	if (*cwd) rawchdir(cwd);
 }
@@ -1332,7 +1373,8 @@ char *argv[], *envp[];
 	}
 #endif	/* !_NOARCHIVE */
 
-	for (i = 0; i < 10; i++) helpindex[i] = strdup2(helpindex[i]);
+	for (i = 0; i < MAXHELPINDEX; i++)
+		helpindex[i] = strdup2(helpindex[i]);
 #ifdef	_USEDOSEMU
 	for (i = 0; fdtype[i].name; i++) {
 		fdtype[i].name = strdup2(fdtype[i].name);
@@ -1350,12 +1392,11 @@ char *argv[], *envp[];
 #endif
 
 	setexecpath(argv[0], envp);
-#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
-# ifdef	DATADIR
+#ifdef	_USEUNICODE
 	unitblpath = strdup2(DATADIR);
-# else
-	unitblpath = strdup2(progpath);
-# endif
+#endif
+#ifndef	_NOIME
+	dicttblpath = strdup2(DATADIR);
 #endif
 
 #ifdef	_NOORIGSHELL
@@ -1372,7 +1413,10 @@ char *argv[], *envp[];
 # endif
 		Xexit2(i);
 	}
-	fdmode = interactive = 1;
+#ifndef	_NOORIGSHELL
+	interactive =
+#endif
+	fdmode = 1;
 	setshellvar(envp);
 	prepareterm();
 	argc = initoption(argc, argv, envp);
@@ -1381,6 +1425,7 @@ char *argv[], *envp[];
 		error(NTERM_K);
 	}
 #endif	/* !_NOORIGSHELL */
+	initenv();
 
 	Xttyiomode(0);
 	initterm();
