@@ -1,7 +1,7 @@
 /*
  *	parse.c
  *
- *	Commandline Parser
+ *	command line parser
  */
 
 #include "fd.h"
@@ -92,7 +92,7 @@ strtable keyidentlist[] = {
 	{K_HELP,	"HELP"},
 	{0,		NULL}
 };
-#define	KEYIDENTSIZ	((int)(sizeof(keyidentlist) / sizeof(strtable)) - 1)
+#define	KEYIDENTSIZ	(arraysize(keyidentlist) - 1)
 
 static CONST char escapechar[] = "abefnrtv";
 static CONST char escapevalue[] = {
@@ -103,7 +103,8 @@ static CONST char escapevalue[] = {
 char *skipspace(cp)
 char *cp;
 {
-	while (*cp == ' ' || *cp == '\t') cp++;
+	while (isblank2(*cp)) cp++;
+
 	return(cp);
 }
 
@@ -114,14 +115,15 @@ int c;
 	else if (isdigit2(c)) return(c - '0');
 	else if (islower2(c)) return(c - 'a' + 10);
 	else if (isupper2(c)) return(c - 'A' + 10);
-	else return(-1);
+
+	return(-1);
 }
 
 #ifdef	USESTDARGH
-/*VARARGS1*/
+/*VARARGS2*/
 char *sscanf2(char *s, CONST char *fmt, ...)
 #else
-/*VARARGS1*/
+/*VARARGS2*/
 char *sscanf2(s, fmt, va_alist)
 char *s;
 CONST char *fmt;
@@ -134,16 +136,15 @@ va_dcl
 	u_long_t u, mask;
 	int i, c, len, base, width, flags;
 
-#ifdef	USESTDARGH
-	va_start(args, fmt);
-#else
-	va_start(args);
-#endif
-
 	if (!s || !fmt) return(NULL);
+	VA_START(args, fmt);
+
 	for (i = 0; fmt[i]; i++) {
 		if (fmt[i] != '%') {
-			if (*(s++) != fmt[i]) return(NULL);
+			if (*(s++) != fmt[i]) {
+				s = NULL;
+				break;
+			}
 			continue;
 		}
 
@@ -184,22 +185,26 @@ va_dcl
 				base = 16;
 				break;
 			case 'c':
-				if (*(s++) != va_arg(args, int)) return(NULL);
+				if (*(s++) != va_arg(args, int)) s = NULL;
 				break;
 			case '$':
-				if (*s) return(NULL);
+				if (*s) s = NULL;
 				break;
 			default:
-				if (*(s++) != fmt[i]) return(NULL);
+				if (*(s++) != fmt[i]) s = NULL;
 				break;
 		}
+		if (!s) break;
 		if (!base) continue;
 
 		cp = s;
 		n = (long_t)0;
 		u = (u_long_t)0;
 		if (*s == '-') {
-			if (!(flags & VF_MINUS)) return(NULL);
+			if (!(flags & VF_MINUS)) {
+				s = NULL;
+				break;
+			}
 			for (s++; *s; s++) {
 #ifdef	LSI_C
 			/* for buggy LSI-C */
@@ -218,7 +223,7 @@ va_dcl
 					n = MINTYPE(long_t);
 				else n = n * base - c;
 			}
-			if ((flags & VF_UNSIGNED) && n) return(NULL);
+			if ((flags & VF_UNSIGNED) && n) s = NULL;
 		}
 		else if (flags & VF_UNSIGNED) {
 			for (; *s; s++) {
@@ -231,7 +236,7 @@ va_dcl
 					u = MAXTYPE(u_long_t);
 				else u = u * base + c;
 			}
-			if ((flags & VF_PLUS) && !u) return(NULL);
+			if ((flags & VF_PLUS) && !u) s = NULL;
 		}
 		else {
 			for (; *s; s++) {
@@ -244,33 +249,72 @@ va_dcl
 					n = MAXTYPE(long_t);
 				else n = n * base + c;
 			}
-			if ((flags & VF_PLUS) && !n) return(NULL);
+			if ((flags & VF_PLUS) && !n) s = NULL;
+		}
+		if (!s || s <= cp || ((flags & VF_ZERO) && width > 0)) {
+			s = NULL;
+			break;
 		}
-		if (s <= cp) return(NULL);
-		if ((flags & VF_ZERO) && width > 0) return(NULL);
 
-		mask = (MAXUTYPE(u_long_t)
-			>> ((sizeof(long_t) - len) * BITSPERBYTE));
+		mask = MAXUTYPE(u_long_t);
+		if (len < (int)sizeof(u_long_t))
+			mask >>= ((int)sizeof(u_long_t) - len) * BITSPERBYTE;
 		if (flags & VF_UNSIGNED) {
-			if (u & ~mask) return(NULL);
+			if (u & ~mask) {
+				u = mask;
+				if (flags & VF_STRICTWIDTH) {
+					s = NULL;
+					break;
+				}
+			}
 		}
 		else {
 			mask >>= 1;
 			if (n >= 0) {
-				if ((u_long_t)n & ~mask) return(NULL);
+				if ((u_long_t)n & ~mask) {
+					n = mask;
+					if (flags & VF_STRICTWIDTH) s = NULL;
+				}
 			}
-			else if (((u_long_t)n & ~mask) != ~mask) return(NULL);
+			else if (((u_long_t)n & ~mask) != ~mask) {
+				n = ~mask;
+				if (flags & VF_STRICTWIDTH) s = NULL;
+			}
+			if (!s) break;
 			memcpy(&u, &n, sizeof(u));
 		}
 
-		if (len == sizeof(u_long_t))
+#ifndef	HAVELONGLONG
+		if (len > (int)sizeof(u_long_t)) {
+			char *buf;
+			u_long_t tmp;
+			int hi;
+
+			hi = 0;
+			if (!(flags & VF_UNSIGNED)) {
+				mask = (MAXUTYPE(u_long_t) >> 1);
+				if (u & ~mask) hi = MAXUTYPE(char);
+			}
+			buf = va_arg(args, char *);
+			memset(buf, hi, len);
+
+			tmp = 0x5a;
+			cp = (char *)(&tmp);
+			if (*cp != 0x5a) buf += len - sizeof(u_long_t);
+			memcpy(buf, (char *)(&u), sizeof(u));
+		}
+		else
+#endif	/* !HAVELONGLONG */
+		if (len == (int)sizeof(u_long_t))
 			*(va_arg(args, u_long_t *)) = u;
 #ifdef	HAVELONGLONG
-		else if (len == sizeof(u_long)) *(va_arg(args, u_long *)) = u;
+		else if (len == (int)sizeof(u_long))
+			*(va_arg(args, u_long *)) = u;
 #endif
-		else if (len == sizeof(u_short))
+		else if (len == (int)sizeof(u_short))
 			*(va_arg(args, u_short *)) = u;
-		else if (len == sizeof(u_char)) *(va_arg(args, u_char *)) = u;
+		else if (len == (int)sizeof(u_char))
+			*(va_arg(args, u_char *)) = u;
 		else *(va_arg(args, u_int *)) = u;
 	}
 	va_end(args);
@@ -290,7 +334,7 @@ int evaldq;
 # if	MSDOS
 		pc = parsechar(cp, -1, '\0', 0, &quote, NULL);
 # else
-		pc = parsechar(cp, -1, '\0', 1, &quote, NULL);
+		pc = parsechar(cp, -1, '\0', EA_BACKQ, &quote, NULL);
 # endif
 		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) continue;
 		else if (pc == PC_SQUOTE || pc == PC_BQUOTE) continue;
@@ -301,7 +345,7 @@ int evaldq;
 			cp++;
 			continue;
 		}
-		else if (pc == PC_META) {
+		else if (pc == PC_ESCAPE) {
 			cp++;
 			if (*cp == PMETA && strchr(c, *cp)) return(cp - 1);
 			continue;
@@ -309,6 +353,7 @@ int evaldq;
 
 		if (strchr(c, *cp)) return(cp);
 	}
+
 	return(NULL);
 }
 
@@ -320,6 +365,7 @@ int c, evaldq;
 
 	tmp[0] = c;
 	tmp[1] = '\0';
+
 	return(strtkbrk(s, tmp, evaldq));
 }
 
@@ -333,6 +379,7 @@ char **strp;
 	if ((tmp = strtkbrk(*strp, " \t", 0))) len = tmp - *strp;
 	else len = strlen(*strp);
 	*strp += len;
+
 	return(strndup2(cp, len));
 }
 
@@ -346,10 +393,11 @@ char *s, ***argvp;
 	cp = skipspace(s);
 	for (i = 0; *cp; i++) {
 		*argvp = (char **)realloc2(*argvp, (i + 2) * sizeof(char *));
-		(*argvp)[i] = evalpath(geteostr(&cp), 0);
+		(*argvp)[i] = evalpath(geteostr(&cp), EA_NOUNIQDELIM);
 		cp = skipspace(cp);
 	}
 	(*argvp)[i] = NULL;
+
 	return(i);
 }
 
@@ -358,6 +406,7 @@ char *s;
 {
 	if (!isidentchar(*s)) return(NULL);
 	for (s++; isidentchar2(*s); s++);
+
 	return(s);
 }
 
@@ -386,7 +435,8 @@ char *argv[];
 		cp = argv[i];
 	}
 	*argcp = i + 1;
-	return(evalpath(strdup2(cp), 0));
+
+	return(evalpath(strdup2(cp), EA_NOUNIQDELIM));
 }
 
 char *evalcomstr(path, delim)
@@ -411,7 +461,7 @@ char *path, *delim;
 		}
 		next = cp + len;
 		if (len) {
-			cp = _evalpath(cp, next, 0, 0);
+			cp = _evalpath(cp, next, EA_NOEVALQ | EA_NOUNIQDELIM);
 # ifndef	_NOKANJIFCONV
 			tmp = newkanjiconv(cp, DEFCODE, getkcode(cp), L_FNAME);
 			if (cp != tmp) free(cp);
@@ -433,8 +483,9 @@ char *path, *delim;
 		}
 	}
 
-	if (!epath) return(strdup2(""));
+	if (!epath) return(strdup2(nullstr));
 	epath[size] = '\0';
+
 	return(epath);
 }
 #endif	/* _NOORIGSHELL */
@@ -456,7 +507,7 @@ int delim;
 		next = strchr(cp, delim);
 		len = (next) ? (next++) - cp : strlen(cp);
 		if (len) {
-			tmp = _evalpath(cp, cp + len, 1, 1);
+			tmp = _evalpath(cp, cp + len, 0);
 			if (!isrootdir(cp)) cp = tmp;
 			else cp = realpath2(tmp, buf, 1);
 			len = strlen(cp);
@@ -473,8 +524,9 @@ int delim;
 		if (next) epath[size++] = delim;
 	}
 
-	if (!epath) return(strdup2(""));
+	if (!epath) return(strdup2(nullstr));
 	epath[size] = '\0';
+
 	return(epath);
 }
 
@@ -507,6 +559,7 @@ char *name;
 	buf[i] = '\0';
 	cp = strdup2(cp);
 	free(buf);
+
 	return(cp);
 }
 #endif	/* !MSDOS || !_NOORIGSHELL */
@@ -516,10 +569,10 @@ VOID adjustpath(VOID_A)
 {
 	char *cp, *path;
 
-	if (!(cp = getconstvar("PATH"))) return;
+	if (!(cp = getconstvar(ENVPATH))) return;
 
 	path = evalpaths(cp, PATHDELIM);
-	if (strpathcmp(path, cp)) setenv2("PATH", path, 1);
+	if (strpathcmp(path, cp)) setenv2(ENVPATH, path, 1);
 	free(path);
 }
 #endif	/* _NOORIGSHELL */
@@ -527,25 +580,20 @@ VOID adjustpath(VOID_A)
 char *includepath(path, plist)
 char *path, *plist;
 {
-	char *cp, *next;
+	char *cp, *tmp, *next;
 	int len;
 
 	if (!plist || !*plist) return(NULL);
 	next = plist;
 	for (cp = next; cp && *cp; cp = next) {
 #ifdef	_USEDOSPATH
-		if (_dospath(cp)) next = strchr(&(cp[2]), PATHDELIM);
-		else
+		if (_dospath(cp)) next += 2;
 #endif
-		next = strchr(cp, PATHDELIM);
+		next = strchr(next, PATHDELIM);
 		len = (next) ? (next++) - cp : strlen(cp);
-		while (len > 1 && cp[len - 1] == _SC_) len--;
-#ifdef	BSPATHDELIM
-		if (onkanji1(cp, len - 1)) len++;
-#endif
-		if (len > 0 && !strnpathcmp(path, cp, len)
-		&& (!path[len] || path[len] == _SC_)) return(path + len);
+		if ((tmp = underpath(path, cp, len))) return(tmp);
 	}
+
 	return(NULL);
 }
 
@@ -582,6 +630,7 @@ u_char *fp, *dp, *wp;
 		else if (*cp && *cp != ',' && *cp != ':') *wp = *(cp++) % 128;
 		else return(NULL);
 	}
+
 	return(cp);
 }
 #endif	/* (FD < 2) && !_NOARCHIVE */
@@ -600,9 +649,9 @@ char **bufp, *prompt;
 	int i, j, k, len, unprint;
 
 #ifdef	_NOORIGSHELL
-	prompt = evalpath(strdup2(prompt), 0);
+	prompt = evalpath(strdup2(prompt), EA_NOUNIQDELIM);
 #else
-	prompt = evalvararg(prompt, 0, 1, '\0', 0, 0);
+	prompt = evalvararg(prompt, '\0', EA_BACKQ | EA_KEEPMETA, 0);
 #endif
 	unprint = 0;
 #ifdef	FAKEUNINIT
@@ -647,7 +696,8 @@ char **bufp, *prompt;
 				gethostname(line, MAXPATHLEN);
 # endif
 				if (prompt[i] == 'h'
-				&& (tmp = strchr(line, '.'))) *tmp = '\0';
+				&& (tmp = strchr(line, '.')))
+					*tmp = '\0';
 				break;
 			case '$':
 				*line = (getuid()) ? '$' : '#';
@@ -670,7 +720,7 @@ char **bufp, *prompt;
 				else cp = tmp;
 				break;
 			case '~':
-				if (underhome(line + 1)) line[0] = '~';
+				if (underhome(&(line[1]))) line[0] = '~';
 				else if (!physical_path || !Xgetwd(line))
 					cp = fullpath;
 				break;
@@ -685,7 +735,7 @@ char **bufp, *prompt;
 				unprint = 0;
 				break;
 			default:
-				tmp = sscanf2(&(prompt[i]), "%3Co", line);
+				tmp = sscanf2(&(prompt[i]), "%<3Co", line);
 				if (tmp) i = (tmp - prompt) - 1;
 				else *line = prompt[i];
 				line[1] = '\0';
@@ -723,6 +773,7 @@ char **bufp, *prompt;
 	}
 	(*bufp)[j] = '\0';
 	free(prompt);
+
 	return(len);
 }
 
@@ -747,6 +798,7 @@ u_char *flagsp;
 		*tmp = '*';
 		strcpy(&(tmp[1]), ext);
 	}
+
 	return(tmp);
 }
 
@@ -764,6 +816,7 @@ int flags2, strict;
 	if ((flags1 & LF_IGNORECASE) || (flags2 & LF_IGNORECASE))
 		return(strcasecmp2(ext1, ext2));
 # endif
+
 	return(strpathcmp(ext1, ext2));
 }
 #endif	/* !_NOARCHIVE */
@@ -775,6 +828,7 @@ int identonly;
 	char *tmp;
 	int i, ch;
 
+	if (!cp || !*cp) return(-1);
 	ch = (*(cp++) & 0xff);
 	if (!*cp) {
 		if (identonly) return(-1);
@@ -819,6 +873,7 @@ int identonly;
 			break;
 	}
 	if (*cp) ch = -1;
+
 	return(ch);
 }
 
@@ -826,7 +881,15 @@ char *getkeysym(c, tenkey)
 int c, tenkey;
 {
 	static char buf[5];
-	int i, j;
+	int i;
+
+	for (i = 0; i < KEYIDENTSIZ; i++)
+		if ((u_short)(c) == keyidentlist[i].no) break;
+	if (i < KEYIDENTSIZ) {
+		if (c > (int)MAXUTYPE(u_char)
+		|| tenkey || c == ' ' || iscntrl2(c))
+			return(keyidentlist[i].str);
+	}
 
 	i = 0;
 	if (c >= K_F(1) && c <= K_F(20))
@@ -837,31 +900,28 @@ int c, tenkey;
 	}
 #ifdef	CODEEUC
 	else if (isekana2(c)) {
-		buf[i++] = C_EKANA;
+		buf[i++] = (char)C_EKANA;
 		buf[i++] = c & 0xff;
 	}
 #endif
 	else if (c > (int)MAXUTYPE(u_char)) {
-		for (j = 0; j < KEYIDENTSIZ; j++)
-			if ((u_short)(c) == keyidentlist[j].no) break;
-		if (j < KEYIDENTSIZ) {
-			if (tenkey || c == ' ' || iscntrl2(c)
-			|| keyidentlist[j].no >= K_MIN)
-				return(keyidentlist[j].str);
-		}
-
 		buf[i++] = '?';
 		buf[i++] = '?';
 	}
 #ifndef	CODEEUC
 	else if (iskana2(c)) buf[i++] = c;
 #endif
-	else if (iscntrl2(c))
-		i = snprintf2(buf, sizeof(buf), "^%c", (c + '@') & 0x7f);
+	else if (iscntrl2(c)) {
+		for (i = 0; escapechar[i]; i++)
+			if (c == escapevalue[i]) break;
+		if (escapechar[i])
+			i = snprintf2(buf, sizeof(buf), "\\%c", escapechar[i]);
+		else i = snprintf2(buf, sizeof(buf), "^%c", (c + '@') & 0x7f);
+	}
 	else if (ismsb(c)) i = snprintf2(buf, sizeof(buf), "\\%03o", c);
 	else buf[i++] = c;
-
 	buf[i] = '\0';
+
 	return(buf);
 }
 
@@ -877,7 +937,7 @@ int evalhat;
 	for (i = j = 0; s[i]; i++, j++) {
 		if (s[i] == '\\') {
 			i++;
-			if ((tmp = sscanf2(&(s[i]), "%3Co", &(cp[j])))) {
+			if ((tmp = sscanf2(&(s[i]), "%<3Co", &(cp[j])))) {
 				i = (tmp - s) - 1;
 				continue;
 			}
@@ -898,6 +958,7 @@ int evalhat;
 	s[j] = '\0';
 	free(cp);
 	if (lenp) *lenp = j;
+
 	return(s);
 }
 
@@ -935,6 +996,7 @@ int len;
 		cp[j++] = s[i];
 	}
 	cp[j] = '\0';
+
 	return(cp);
 }
 #endif	/* !_NOKEYMAP */
