@@ -7,13 +7,10 @@
 #ifdef	FD
 #include "fd.h"
 #else	/* !FD */
+#include "machine.h"
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include "machine.h"
-#include "printf.h"
-#include "kctype.h"
-#include "pathname.h"
 
 # ifndef	NOUNISTDH
 # include <unistd.h>
@@ -22,6 +19,10 @@
 # ifndef	NOSTDLIBH
 # include <stdlib.h>
 # endif
+
+#include "printf.h"
+#include "kctype.h"
+#include "pathname.h"
 #endif	/* !FD */
 
 #if	MSDOS
@@ -202,6 +203,9 @@ static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
 #define	getconstvar(s)	(getenvvar(s, sizeof(s) - 1))
+#ifdef	FD
+#define	STRFD		"FD"
+#endif
 
 char nullstr[] = "";
 char rootpath[] = _SS_;
@@ -532,13 +536,25 @@ char *path;
 #endif	/* !FD */
 
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+char *gendospath(path, drive, c)
+char *path;
+int drive, c;
+{
+	*(path++) = drive;
+	*(path++) = ':';
+	if (c) *(path++) = c;
+	*path = '\0';
+
+	return(path);
+}
+
 char *strdelim(s, d)
 char *s;
 int d;
 {
 	int i;
 
-	if (d && _dospath(s)) return(s + 1);
+	if (d && _dospath(s)) return(&(s[1]));
 	for (i = 0; s[i]; i++) {
 		if (s[i] == _SC_) return(&(s[i]));
 # ifdef	BSPATHDELIM
@@ -556,7 +572,7 @@ int d;
 	char *cp;
 	int i;
 
-	if (d && _dospath(s)) cp = s + 1;
+	if (d && _dospath(s)) cp = &(s[1]);
 	else cp = NULL;
 	for (i = 0; s[i]; i++) {
 		if (s[i] == _SC_) cp = &(s[i]);
@@ -621,6 +637,10 @@ char *s;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 	if (_dospath(s)) i = 2;
 	else
+#endif
+#ifdef	DOUBLESLASH
+	if (isdslash(s)) i = 1;
+	else
 #endif
 	i = 0;
 	if (!s[i]) return(&(s[i]));
@@ -660,6 +680,13 @@ char *buf, *s1, *s2;
 		i = 2;
 	}
 	else
+#endif
+#ifdef	DOUBLESLASH
+	if (isdslash(s1)) {
+		buf[0] = s1[0];
+		i = 1;
+	}
+	else
 #endif
 	i = 0;
 	if (!s1[i]) cp = &(buf[i]);
@@ -704,13 +731,17 @@ char *buf, *s1, *s2;
 int strcasecmp2(s1, s2)
 char *s1, *s2;
 {
+	int c1, c2;
+
 	for (;;) {
-		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
+		c1 = toupper2(*s1);
+		c2 = toupper2(*s2);
+		if (c1 != c2) return(c1 - c2);
 #ifndef	CODEEUC
 		if (issjis1(*s1)) {
 			s1++;
 			s2++;
-			if (*s1 != *s2) return(*s1 - *s2);
+			if (*s1 != *s2) return((u_char)*s1 - (u_char)*s2);
 		}
 #endif
 		if (!*s1) break;
@@ -725,14 +756,18 @@ int strncasecmp2(s1, s2, n)
 char *s1, *s2;
 int n;
 {
+	int c1, c2;
+
 	while (n-- > 0) {
-		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
+		c1 = toupper2(*s1);
+		c2 = toupper2(*s2);
+		if (c1 != c2) return(c1 - c2);
 #ifndef	CODEEUC
 		if (issjis1(*s1)) {
 			if (n-- <= 0) break;
 			s1++;
 			s2++;
-			if (*s1 != *s2) return(*s1 - *s2);
+			if (*s1 != *s2) return((u_char)*s1 - (u_char)*s2);
 		}
 #endif
 		if (!*s1) break;
@@ -769,7 +804,7 @@ int len;
 	char *cp;
 
 	if (len < 0) len = strlen(dir);
-	if ((cp = strrdelim2(dir, &(dir[len]))) && !*(cp + 1)) len = cp - dir;
+	if ((cp = strrdelim2(dir, &(dir[len]))) && !cp[1]) len = cp - dir;
 	if (len <= 0 || strnpathcmp(path, dir, len)) return(NULL);
 	if (path[len] && path[len] != _SC_) return(NULL);
 
@@ -857,7 +892,7 @@ char *s;
 int isrootpath(s)
 char *s;
 {
-	return((s[0] == _SC_ && !s[1]));
+	return((s[0] == _SC_ && !s[1]) ? 1 : 0);
 }
 
 VOID copyrootpath(s)
@@ -874,12 +909,29 @@ char *s;
 	*s = '\0';
 }
 
+#ifdef	DOUBLESLASH
+int isdslash(s)
+char *s;
+{
+# if	MSDOS || defined (CYGWIN)
+	char *cp;
+
+	if (s[0] != _SC_ || s[1] != _SC_ || !s[2] || s[2] == _SC_) return(0);
+	if ((cp = strdelim(&(s[2]), 0))) return(cp - s);
+
+	return(strlen(s));
+# else
+	return((s[0] == _SC_ && s[1] == _SC_) ? 2 : 0);
+# endif
+}
+#endif	/* DOUBLESLASH */
+
 char *getbasename(s)
 char *s;
 {
 	char *cp;
 
-	if ((cp = strrdelim(s, 1))) return(cp + 1);
+	if ((cp = strrdelim(s, 1))) return(&(cp[1]));
 
 	return(s);
 }
@@ -1404,7 +1456,7 @@ int len;
 static void NEAR duplwild(dst, src)
 wild_t *dst, *src;
 {
-	memcpy(dst, src, sizeof(wild_t));
+	memcpy((char *)dst, (char *)src, sizeof(wild_t));
 	dst -> fixed.s = (char *)malloc2(src -> fixed.size);
 	memcpy(dst -> fixed.s, src -> fixed.s, src -> fixed.len + 1);
 	dst -> path.s = (char *)malloc2(src -> path.size);
@@ -1414,7 +1466,8 @@ wild_t *dst, *src;
 	if (src -> ino) {
 		dst -> ino = (devino_t *)malloc2(src -> nino
 			* sizeof(devino_t));
-		memcpy(dst -> ino, src -> ino, src -> nino * sizeof(devino_t));
+		memcpy((char *)(dst -> ino), (char *)(src -> ino),
+			src -> nino * sizeof(devino_t));
 	}
 #endif	/* !NODIRLOOP */
 }
@@ -1434,6 +1487,9 @@ int argc;
 char ***argvp;
 wild_t *wp;
 {
+#ifdef	DOUBLESLASH
+	int ds;
+#endif
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
@@ -1457,6 +1513,14 @@ wild_t *wp;
 	quote = wp -> quote;
 
 	if (wp -> fixed.len) addstrbuf(&(wp -> path), rootpath, 1);
+#ifdef	DOUBLESLASH
+	if (wp -> path.len) ds = 0;
+	else if ((ds = isdslash(wp -> s))) {
+		addstrbuf(&(wp -> fixed), wp -> s, ds);
+		addstrbuf(&(wp -> path), wp -> s, ds);
+		wp -> s += ds;
+	}
+#endif
 
 	for (i = w = 0; wp -> s[i] && wp -> s[i] != _SC_; i++) {
 		pc = parsechar(&(wp -> s[i]), -1,
@@ -1503,6 +1567,9 @@ wild_t *wp;
 
 	if (!w) {
 		if (wp -> path.len <= plen) w++;
+#ifdef	DOUBLESLASH
+		else if (ds) st.st_mode = S_IFDIR;
+#endif
 		else if (stat2(wp -> path.s, &st) < 0) return(argc);
 
 		wp -> s += i;
@@ -1513,13 +1580,17 @@ wild_t *wp;
 		}
 
 #ifndef	NODIRLOOP
+# ifdef	DOUBLESLASH
+		if (ds) /*EMPTY*/;
+		else
+# endif
 		if (!w) {
 			wp -> ino = (devino_t *)realloc2(wp -> ino,
 				(wp -> nino + 1) * sizeof(devino_t));
 			wp -> ino[wp -> nino].dev = st.st_dev;
 			wp -> ino[(wp -> nino)++].ino = st.st_ino;
 		}
-#endif
+#endif	/* !NODIRLOOP */
 		return(_evalwild(argc, argvp, wp));
 	}
 
@@ -1794,18 +1865,18 @@ int len, exe;
 {
 	if (ext) {
 		if (isexecute(path, 0, exe) >= 0) {
-			if (!(strpathcmp(ext, "COM"))) return(0);
-			else if (!(strpathcmp(ext, "EXE"))) return(CM_EXE);
-			else if (!(strpathcmp(ext, "BAT"))) return(CM_BATCH);
+			if (!(strpathcmp(ext, EXTCOM))) return(0);
+			else if (!(strpathcmp(ext, EXTEXE))) return(CM_EXE);
+			else if (!(strpathcmp(ext, EXTBAT))) return(CM_BATCH);
 		}
 	}
 	else {
 		path[len++] = '.';
-		strcpy(&(path[len]), "COM");
+		strcpy(&(path[len]), EXTCOM);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT);
-		strcpy(&(path[len]), "EXE");
+		strcpy(&(path[len]), EXTEXE);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_EXE);
-		strcpy(&(path[len]), "BAT");
+		strcpy(&(path[len]), EXTBAT);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_BATCH);
 	}
 
@@ -1893,7 +1964,7 @@ char *com, *search;
 #ifndef	_NOUSEHASH
 	recalc = 0;
 #endif
-	if ((next = (search) ? search : getconstvar("PATH"))) {
+	if ((next = (search) ? search : getconstvar(ENVPATH))) {
 		len = strlen(com);
 		size = ret = 0;
 		path = NULL;
@@ -1910,7 +1981,7 @@ char *com, *search;
 			dlen = (next) ? (next++) - cp : strlen(cp);
 			if (!dlen) tmp = NULL;
 			else {
-				tmp = _evalpath(cp, cp + dlen, 0);
+				tmp = _evalpath(cp, &(cp[dlen]), 0);
 				dlen = strlen(tmp);
 			}
 			if (dlen + len + 1 + EXTWIDTH + 1 > size) {
@@ -2177,7 +2248,7 @@ char ***argvp;
 # ifdef	CWDINPATH
 	argc = completefile(file, len, argc, argvp, curpath, 1, 2);
 # endif
-	if (!(next = getconstvar("PATH"))) return(argc);
+	if (!(next = getconstvar(ENVPATH))) return(argc);
 	for (cp = next; cp; cp = next) {
 # if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 		if (_dospath(cp)) next = strchr(&(cp[2]), PATHDELIM);
@@ -2185,7 +2256,7 @@ char ***argvp;
 # endif
 		next = strchr(cp, PATHDELIM);
 		dlen = (next) ? (next++) - cp : strlen(cp);
-		tmp = _evalpath(cp, cp + dlen, 0);
+		tmp = _evalpath(cp, &(cp[dlen]), 0);
 		dlen = strlen(tmp);
 		argc = completefile(file, len, argc, argvp, tmp, dlen, 2);
 		free(tmp);
@@ -2200,6 +2271,12 @@ int len, argc;
 char ***argvp;
 int exe;
 {
+# if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	char cwd[4];
+# endif
+# ifdef	DOUBLESLASH
+	int n;
+# endif
 	char *file, *dir;
 	int dlen;
 
@@ -2213,16 +2290,34 @@ int exe;
 # endif
 
 	if ((file = strrdelim(dir, 0))) {
-		dlen += (file == dir) ? 1 : file - dir;
-		return(completefile(file + 1, strlen(file + 1), argc, argvp,
+		if (file == dir) {
+			dlen++;
+			file++;
+		}
+# ifdef	DOUBLESLASH
+		else if ((n = isdslash(path)) && file < &(path[n])) {
+			if (path[n - 1] != _SC_) return(argc);
+			dlen = n;
+			file = &(path[n]);
+			if (*file == _SC_) file++;
+		}
+# endif
+		else dlen += file++ - dir;
+		return(completefile(file, strlen(file), argc, argvp,
 			path, dlen, exe));
 	}
 # ifndef	NOUID
 	else if (*path == '~')
 		return(completeuser(&(path[1]), len - 1, argc, argvp, 1));
 # endif
-	else if (exe && dir == path)
-		return(completeexe(path, len, argc, argvp));
+# if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	else if (dir != path) {
+		len -= dlen;
+		dlen = gendospath(cwd, *path, '.') - cwd;
+		return(completefile(dir, len, argc, argvp, cwd, dlen, exe));
+	}
+# endif
+	else if (exe) return(completeexe(path, len, argc, argvp));
 
 	return(completefile(path, len, argc, argvp, curpath, 1, exe));
 }
@@ -2579,13 +2674,13 @@ int plen, mode;
 	ret = NULL;
 	len = strlen(s);
 	if ((mode & ~0x80) != '#') {
-		if (mode & 0x80) for (cp = s; cp < s + len; cp++) {
+		if (mode & 0x80) for (cp = s; cp < &(s[len]); cp++) {
 			if (regexp_exec(re, cp, 0)) {
 				ret = cp;
 				break;
 			}
 		}
-		else for (cp = s + len - 1; cp >= s; cp--) {
+		else for (cp = &(s[len - 1]); cp >= s; cp--) {
 			if (regexp_exec(re, cp, 0)) {
 				ret = cp;
 				break;
@@ -2679,7 +2774,7 @@ int plen, *modep;
 			}
 #ifdef	BASHSTYLE
 	/* bash uses IFS instead of a space as a separator */
-			if ((cp = getconstvar("IFS")) && *cp)
+			if ((cp = getconstvar(ENVIFS)) && *cp)
 				sp = *cp;
 #endif
 
@@ -2713,7 +2808,7 @@ int plen, *modep;
 			sp = ' ';
 #ifdef	BASHSTYLE
 	/* bash uses IFS instead of a space as a separator */
-			if (quoted && (cp = getconstvar("IFS")))
+			if (quoted && (cp = getconstvar(ENVIFS)))
 				sp = (*cp) ? *cp : '\0';
 #endif
 			cp = catvar(&(arglist[1]), sp);
@@ -3143,7 +3238,7 @@ char *gethomedir(VOID_A)
 #endif	/* !NOUID */
 	char *cp;
 
-	if ((cp = getconstvar("HOME"))) return(cp);
+	if ((cp = getconstvar(ENVHOME))) return(cp);
 #ifndef	NOUID
 # ifdef	FD
 	if ((up = finduid(getuid(), NULL))) return(up -> home);
@@ -3180,7 +3275,7 @@ char **argp;
 	len = ((cp = strdelim(top, 0))) ? (cp - top) : strlen(top);
 	if (!len) cp = gethomedir();
 # ifdef	FD
-	else if (len == strsize("FD") && !strnpathcmp(top, "FD", len))
+	else if (len == strsize(STRFD) && !strnpathcmp(top, STRFD, len))
 		cp = progpath;
 # endif
 	else {
@@ -3289,7 +3384,7 @@ int qed, flags;
 			if (*cp == '`') {
 				bbuf[j] = '\0';
 				buf = replacebackquote(buf, &i,
-					bbuf, strlen(cp + 1));
+					bbuf, strlen(&(cp[1])));
 				j = 0;
 			}
 			else if (!(flags & EA_STRIPQ)) buf[i++] = *cp;
@@ -3495,7 +3590,10 @@ int flags;
 {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char alias[MAXPATHLEN];
-	int top = -1;
+	int top;
+#endif
+#ifdef	DOUBLESLASH
+	int ds;
 #endif
 	char *cp, *tmp;
 	int i, j, c, pc, size, quote;
@@ -3511,9 +3609,16 @@ int flags;
 	free(cp);
 	cp = tmp;
 
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
+	top = -1;
+#endif
+#ifdef	DOUBLESLASH
+	ds = isdslash(cp);
+#endif
 	size = strlen(cp) + 1;
 	tmp = malloc2(size);
 	quote = '\0';
+
 	for (i = j = c = 0; cp[i]; c = cp[i++]) {
 		pc = parsechar(&(cp[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_CLQUOTE) {
@@ -3548,9 +3653,12 @@ int flags;
 			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
-		else if (!(flags & EA_NOUNIQDELIM)
-		&& cp[i] == _SC_ && c == _SC_)
-			continue;
+		else if (flags & EA_NOUNIQDELIM) /*EMPTY*/;
+#ifdef	DOUBLESLASH
+		else if (ds && i == 1) /*EMPTY*/;
+#endif
+		else if (cp[i] == _SC_ && c == _SC_) continue;
+
 		tmp[j++] = cp[i];
 	}
 	tmp[j] = '\0';
