@@ -21,6 +21,10 @@
 #define	NESTINGQUOTE
 #endif
 
+#if	defined (MINIMUMSHELL) && defined (DOUBLESLASH)
+#undef	DOUBLESLASH
+#endif
+
 #define	IFS_SET		" \t\n"
 #define	META		'\\'
 #if	MSDOS && defined (_NOORIGSHELL)
@@ -29,6 +33,15 @@
 #define	PATHDELIM	':'
 #endif
 
+#define	ENVPATH		"PATH"
+#define	ENVHOME		"HOME"
+#define	ENVPWD		"PWD"
+#define	ENVIFS		"IFS"
+
+#define	EXTCOM		"COM"
+#define	EXTEXE		"EXE"
+#define	EXTBAT		"BAT"
+
 #ifndef	BSPATHDELIM
 #define	PMETA		META
 #define	RMSUFFIX	'%'
@@ -196,6 +209,7 @@ typedef struct _gidtable {
 } gidtable;
 
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+extern char *gendospath __P_((char *, int, int));
 extern char *strdelim __P_((char *, int));
 extern char *strrdelim __P_((char *, int));
 #else
@@ -241,6 +255,9 @@ extern char *isrootdir __P_((char *));
 extern int isrootpath __P_((char *));
 extern VOID copyrootpath __P_((char *));
 extern VOID copycurpath __P_((char *));
+#ifdef	DOUBLESLASH
+extern int isdslash __P_((char *));
+#endif
 extern char *getbasename __P_((char *));
 extern char *getshellname __P_((char *, int *, int *));
 extern reg_t *regexp_init __P_((char *, int));
