@@ -8,11 +8,10 @@
 #ifdef	FD
 #include "fd.h"
 #else	/* !FD */
+#include "machine.h"
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include "machine.h"
-#include "pathname.h"
 
 #ifndef	NOUNISTDH
 #include <unistd.h>
@@ -21,17 +20,17 @@
 #ifndef	NOSTDLIBH
 #include <stdlib.h>
 #endif
+
+#include "printf.h"
+#include "kctype.h"
+#include "pathname.h"
 #endif	/* !FD */
 
-#include "kctype.h"
-
 #if	MSDOS
-#include <io.h>
 #include "unixemu.h"
-#else	/* !MSDOS */
+#else
 #include <sys/param.h>
-#include <sys/ioctl.h>
-#endif	/* !MSDOS */
+#endif
 
 #include "system.h"
 
@@ -43,28 +42,33 @@
 extern int Xstat __P_((char *, struct stat *));
 extern int Xlstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
-extern int kanjifputs __P_((char *, FILE *));
+# ifndef	_NOPTY
+#include "termemu.h"
+extern VOID sendparent __P_((int, ...));
+# endif
 #else	/* !FD */
 extern int ttyio;
 extern FILE *ttyout;
-#define	Xstat(f, s)	(stat(f, s) ? -1 : 0)
 # if	MSDOS
-# define	Xlstat(f, s)	(stat(f, s) ? -1 : 0)
+extern int Xstat __P_((char *, struct stat *));
+# define	Xlstat		Xstat
 # else
-# define	Xlstat		lstat
+# define	Xstat(p, s)	((stat(p, s)) ? -1 : 0)
+# define	Xlstat(p, s)	((lstat(p, s)) ? -1 : 0)
 # endif
-#define	Xaccess(p, m)	(access(p, m) ? -1 : 0)
-#define	kanjifputs	fputs
+#define	Xaccess(p, m)	((access(p, m)) ? -1 : 0)
 #endif	/* !FD */
 
 #ifndef	EPERM
 #define	EPERM		EACCES
 #endif
 
+extern VOID error __P_((char *));
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *strdup2 __P_((char *));
 extern char *strndup2 __P_((char *, int));
+extern char *strchr2 __P_((char *, int));
 extern char *strncpy2 __P_((char *, char *, int));
 extern int setenv2 __P_((char *, char *, int));
 extern time_t time2 __P_((VOID_A));
@@ -73,6 +77,7 @@ extern time_t time2 __P_((VOID_A));
 static char *NEAR headstree __P_((syntaxtree *));
 #endif
 #ifndef	NOALIAS
+static int NEAR searchalias __P_((char *, int));
 static shaliastable *NEAR duplalias __P_((shaliastable *));
 static int cmpalias __P_((CONST VOID_P, CONST VOID_P));
 #endif
@@ -87,16 +92,17 @@ static int NEAR testsub3 __P_((int, char *[], int *, int));
 
 #define	ALIASDELIMIT	"();&|<>"
 #define	BUFUNIT		32
-#define	getconstvar(s)	(getshellvar(s, sizeof(s) - 1))
+#define	getconstvar(s)	(getshellvar(s, strsize(s)))
 #define	ER_COMNOFOUND	1
 #define	ER_NOTFOUND	2
 #define	ER_NOTIDENT	4
 #define	ER_BADOPTIONS	13
-#define	ER_MISSARG	15
-#define	ER_NOTALIAS	19
-#define	ER_NOSUCHJOB	20
-#define	ER_NUMOUTRANGE	21
-#define	ER_UNKNOWNSIG	22
+#define	ER_NUMOUTRANGE	18
+#define	ER_NOTALIAS	20
+#define	ER_MISSARG	21
+#define	ER_NOSUCHJOB	22
+#define	ER_TERMINATED	23
+#define	ER_UNKNOWNSIG	26
 
 #if	!MSDOS
 typedef struct _mailpath_t {
@@ -107,14 +113,14 @@ typedef struct _mailpath_t {
 #endif
 
 #ifndef	NOJOB
-int gettermio __P_((p_id_t));
+int gettermio __P_((p_id_t, int));
 VOID dispjob __P_((int, FILE *));
 int searchjob __P_((p_id_t, int *));
 int getjob __P_((char *));
 int stackjob __P_((p_id_t, int, syntaxtree *));
 int stoppedjob __P_((p_id_t));
 VOID killjob __P_((VOID_A));
-VOID checkjob __P_((int));
+VOID checkjob __P_((FILE *));
 #endif	/* !NOJOB */
 char *evalposixsubst __P_((char *, int *));
 #if	!MSDOS
@@ -122,6 +128,8 @@ VOID replacemailpath __P_((char *, int));
 VOID checkmail __P_((int));
 #endif
 #ifndef	NOALIAS
+int addalias __P_((char *, char *));
+int deletealias __P_((char *));
 VOID freealias __P_((shaliastable *));
 int checkalias __P_((syntaxtree *, char *, int, int));
 #endif
@@ -162,14 +170,16 @@ static int maxmailpath = 0;
 
 
 #ifndef	NOJOB
-int gettermio(pgrp)
+int gettermio(pgrp, job)
 p_id_t pgrp;
+int job;
 {
 	int ret;
 	sigmask_t mask, omask;
 
-	if (ttypgrp < 0L || pgrp < 0L || pgrp == ttypgrp) return(0);
-	else if (!jobok) {
+	if (ttypgrp < (p_id_t)0 || pgrp < (p_id_t)0 || pgrp == ttypgrp)
+		return(0);
+	else if (!job) {
 		ttypgrp = pgrp;
 		return(0);
 	}
@@ -190,17 +200,12 @@ p_id_t pgrp;
 	Xsigblock(omask, mask);
 
 #ifdef	JOBVERBOSE
-	fputs("gettermio: ", ttyout);
-	fputlong(mypid, ttyout);
-	fputs(": ", ttyout);
-	fputlong(ttypgrp, ttyout);
-	fputs(" -> ", ttyout);
-	fputlong(pgrp, ttyout);
-	fputc('\n', ttyout);
-	fflush(ttyout);
+	fprintf2(ttyout, "gettermio: %id: %id -> %id", mypid, ttypgrp, pgrp);
+	fputnl(ttyout);
 #endif	/* JOBVERBOSE */
 	if ((ret = settcpgrp(ttyio, pgrp)) >= 0) ttypgrp = pgrp;
 	Xsigsetmask(omask);
+
 	return(ret);
 }
 
@@ -220,6 +225,7 @@ syntaxtree *trp;
 	if (!trp || !(trp -> comm)) return(NULL);
 	id = (trp -> comm) -> id;
 	if (id == SM_CHILD || id == SM_STATEMENT) return(NULL);
+
 	return(statementlist[id - 1].ident);
 }
 
@@ -230,20 +236,14 @@ FILE *fp;
 	int i, sig;
 
 	if (n < 0 || n >= maxjobs || !(joblist[n].pids)) return;
-	fputc('[', fp);
-	fputlong(n + 1, fp);
-	fputc(']', fp);
-	if (n == lastjob) fputc('+', fp);
-	else if (n == prevjob) fputc('-', fp);
-	else fputc(' ', fp);
-	fputc(' ', fp);
 	i = joblist[n].npipe;
-	fputlong(joblist[n].pids[i], fp);
-	fputc(' ', fp);
+	fprintf2(fp, "[%d]%c %id ",
+		n + 1, (n == lastjob) ? '+' : ((n == prevjob) ? '-' : ' '),
+		joblist[n].pids[i]);
 	sig = joblist[n].stats[i];
 
-	if (!sig) fputstr("Running", 28, fp);
-	else if (sig < 0) fputstr("Done", 28, fp);
+	if (sig <= 0)
+		fprintf2(fp, "%-28.28s", (sig) ? "Done" : "Running");
 	else {
 		if (sig >= 128) sig -= 128;
 		dispsignal(sig, 28, fp);
@@ -254,8 +254,7 @@ FILE *fp;
 			joblist[n].trp -> type = OP_NONE;
 		printstree(joblist[n].trp, -1, fp);
 	}
-	fputc('\n', fp);
-	fflush(fp);
+	fputnl(fp);
 }
 
 int searchjob(pid, np)
@@ -272,6 +271,7 @@ int *np;
 			return(i);
 		}
 	}
+
 	return(-1);
 }
 
@@ -302,7 +302,7 @@ char *s;
 	if (i < 0 || i >= maxjobs || !joblist || !(joblist[i].pids))
 		return(-1);
 	j = joblist[i].npipe;
-	if (joblist[i].stats[j] < 0 || joblist[i].stats[j] >= 128) return(-1);
+	if (joblist[i].stats[j] < 0 || joblist[i].stats[j] >= 128) return(-2);
 
 	return(i);
 }
@@ -384,20 +384,12 @@ syntaxtree *trp;
 	}
 
 #ifdef	JOBVERBOSE
-	fputs("stackjob: ", ttyout);
-	fputlong(mypid, ttyout);
-	fputs(": ", ttyout);
-	fputlong(pid, ttyout);
-	fputs(", ", ttyout);
-	fputlong(i, ttyout);
-	fputc(':', ttyout);
-	for (j = 0; j <= joblist[i].npipe; j++) {
-		fputlong(joblist[i].pids[j], ttyout);
-		fputc(' ', ttyout);
-	}
-	fputc('\n', ttyout);
-	fflush(ttyout);
+	fprintf2(ttyout, "stackjob: %id: %id, %d:", mypid, pid, i);
+	for (j = 0; j <= joblist[i].npipe; j++)
+		fprintf2(ttyout, "%id ", joblist[i].pids[j]);
+	fputnl(ttyout);
 #endif	/* JOBVERBOSE */
+
 	return(i);
 }
 
@@ -414,16 +406,17 @@ p_id_t pid;
 			else if (!sig) return(0);
 		}
 	}
+
 	return(-1);
 }
 
-VOID checkjob(verbose)
-int verbose;
+VOID checkjob(fp)
+FILE *fp;
 {
 	int i, j;
 
 	while (waitjob(-1, NULL, WNOHANG | WUNTRACED) > 0);
-	if (verbose) for (i = 0; i < maxjobs; i++) {
+	if (fp) for (i = 0; i < maxjobs; i++) {
 		if (!(joblist[i].pids)) continue;
 		j = joblist[i].npipe;
 		if (joblist[i].stats[j] >= 0 && joblist[i].stats[j] < 128)
@@ -432,7 +425,7 @@ int verbose;
 		if (joblist[i].trp
 		&& (isopbg(joblist[i].trp) || isopnown(joblist[i].trp)))
 			joblist[i].trp -> type = OP_NONE;
-		if (jobok && interactive && !nottyout) dispjob(i, stderr);
+		if (jobok && interactive && !nottyout) dispjob(i, fp);
 		free(joblist[i].pids);
 		free(joblist[i].stats);
 		if (joblist[i].trp) {
@@ -462,6 +455,76 @@ VOID killjob(VOID_A)
 #endif	/* !NOJOB */
 
 #ifndef	NOALIAS
+static int NEAR searchalias(ident, len)
+char *ident;
+int len;
+{
+	int i;
+
+	if (len < 0) len = strlen(ident);
+	for (i = 0; shellalias[i].ident; i++)
+		if (!strncommcmp(ident, shellalias[i].ident, len)
+		&& !(shellalias[i].ident[len]))
+			break;
+
+	return(i);
+}
+
+int addalias(ident, comm)
+char *ident, *comm;
+{
+	int i;
+
+	i = searchalias(ident, -1);
+	if (shellalias[i].ident) {
+		free(shellalias[i].ident);
+		free(shellalias[i].comm);
+	}
+	else {
+		shellalias = (shaliastable *)realloc2(shellalias,
+			(i + 2) * sizeof(shaliastable));
+		shellalias[i + 1].ident = NULL;
+	}
+
+	shellalias[i].ident = ident;
+	shellalias[i].comm = comm;
+# if	defined (FD) && !defined (_NOPTY)
+	sendparent(TE_ADDALIAS, ident, comm);
+# endif
+
+	return(0);
+}
+
+int deletealias(ident)
+char *ident;
+{
+	reg_t *re;
+	int i, n, max;
+
+	n = 0;
+	re = regexp_init(ident, -1);
+	for (max = 0; shellalias[max].ident; max++);
+	for (i = 0; i < max; i++) {
+		if (re) {
+			if (!regexp_exec(re, shellalias[i].ident, 0)) continue;
+		}
+		else if (strcommcmp(ident, shellalias[i].ident)) continue;
+		free(shellalias[i].ident);
+		free(shellalias[i].comm);
+		memmove((char *)&(shellalias[i]), (char *)&(shellalias[i + 1]),
+			(max-- - i) * sizeof(shaliastable));
+		i--;
+		n++;
+	}
+	if (re) regexp_free(re);
+	if (!n) return(-1);
+# if	defined (FD) && !defined (_NOPTY)
+	sendparent(TE_DELETEALIAS, ident);
+# endif
+
+	return(0);
+}
+
 static shaliastable *NEAR duplalias(alias)
 shaliastable *alias;
 {
@@ -476,6 +539,7 @@ shaliastable *alias;
 		dupl[i].comm = strdup2(alias[i].comm);
 	}
 	dupl[i].ident = NULL;
+
 	return(dupl);
 }
 
@@ -501,6 +565,7 @@ CONST VOID_P vp2;
 
 	ap1 = (shaliastable *)vp1;
 	ap2 = (shaliastable *)vp2;
+
 	return(strpathcmp2(ap1 -> ident, ap2 -> ident));
 }
 
@@ -514,18 +579,18 @@ int len, delim;
 	if (!trp || (trp -> flags & ST_NODE) || hascomm(trp)
 	|| !interactive
 	|| ((i = getstatid(parentstree(trp))) >= 0
-	&& !(statementlist[i].type & STT_NEEDLIST))) return(-1);
+	&& !(statementlist[i].type & STT_NEEDLIST)))
+		return(-1);
 
 	if ((!strchr(IFS_SET, delim) && !strchr(ALIASDELIMIT, delim)))
 		return(-1);
-	for (i = 0; shellalias[i].ident; i++)
-		if (!(shellalias[i].flags & AL_USED)
-		&& !strncommcmp(ident, shellalias[i].ident, len)
-		&& !(shellalias[i].ident[len])) break;
-	if (!(shellalias[i].ident)) return(-1);
+	i = searchalias(ident, len);
+	if (!(shellalias[i].ident) || (shellalias[i].flags & AL_USED))
+		return(-1);
 
 	*ident = '\0';
 	shellalias[i].flags |= AL_USED;
+
 	return(i);
 }
 #endif	/* NOALIAS */
@@ -536,60 +601,67 @@ int ptr;
 long *resultp;
 {
 	long n;
+	int i, c, base;
 
 	while (s[ptr] && strchr(IFS_SET, s[ptr])) ptr++;
 	if (!s[ptr]) return(-1);
 
-	if (s[ptr] == '+') ptr = _evalexpression(s, ptr + 1, resultp);
+	if (s[ptr] == '+') {
+		if ((ptr = _evalexpression(s, ptr + 1, resultp)) < 0)
+			return(-1);
+	}
 	else if (s[ptr] == '-') {
-		ptr = _evalexpression(s, ptr + 1, resultp);
+		if ((ptr = _evalexpression(s, ptr + 1, resultp)) < 0)
+			return(-1);
 		*resultp = -(*resultp);
 	}
 	else if (s[ptr] == '~') {
-		ptr = _evalexpression(s, ptr + 1, resultp);
+		if ((ptr = _evalexpression(s, ptr + 1, resultp)) < 0)
+			return(-1);
 		*resultp = ~(*resultp);
 	}
 	else if (s[ptr] == '!' && s[ptr + 1] != '=') {
-		ptr = _evalexpression(s, ptr + 1, resultp);
+		if ((ptr = _evalexpression(s, ptr + 1, resultp)) < 0)
+			return(-1);
 		*resultp = (*resultp) ? 0L : 1L;
 	}
 	else if (s[ptr] == '(') {
-		ptr = evalexpression(s, ptr + 1, resultp, 9);
-		if (ptr >= 0) {
-			while (strchr(IFS_SET, s[ptr])) ptr++;
-			if (s[ptr] != ')') ptr = -1;
-			else ptr++;
-		}
+		if ((ptr = evalexpression(s, ptr + 1, resultp, 9)) < 0)
+			return(-1);
+		while (strchr(IFS_SET, s[ptr])) ptr++;
+		if (s[ptr++] != ')') return(-1);
 	}
-	else if (s[ptr] == '0') {
+	else if (isdigit2(s[ptr])) {
+		if (s[ptr] != '0') base = 10;
+		else if (toupper2(s[++ptr]) != 'X') base = 8;
+		else {
+			base = 16;
+			ptr++;
+		}
+
 		n = 0;
-		if (toupper2(s[++ptr]) == 'X') for (ptr++; s[ptr]; ptr++) {
-			if (isdigit(s[ptr])) n = n * 16L + s[ptr] - '0';
-			else if (s[ptr] >= 'a' && s[ptr] <= 'f')
-				n = n * 16L + s[ptr] - 'a' + 10;
-			else if (s[ptr] >= 'A' && s[ptr] <= 'F')
-				n = n * 16L + s[ptr] - 'A' + 10;
-			else break;
-			if (n < 0L) return(-1);
-		}
-		else for (; s[ptr]; ptr++) {
-			if (s[ptr] < '0' || s[ptr] > '7') break;
-			n = n * 8L + s[ptr] - '0';
-			if (n < 0L) return(-1);
+		for (i = ptr; s[i]; i++) {
+			c = s[i];
+			if (isdigit2(c)) c -= '0';
+			else if (islower2(c)) c -= 'a' - 10;
+			else if (isupper2(c)) c -= 'A' - 10;
+			else c = -1;
+			if (c < 0 || c >= base) break;
+
+			if (n > MAXTYPE(long) / base
+			|| (n == MAXTYPE(long) / base
+			&& c > MAXTYPE(long) % base))
+				return(-1);
+			n = n * base + c;
 		}
+		if (i <= ptr) return(-1);
+		ptr = i;
 		*resultp = n;
 	}
-	else if (isdigit(s[ptr])) {
-		n = (long)(s[ptr++] - '0');
-		while (isdigit(s[ptr])) {
-			n = n * 10L + (s[ptr++] - '0');
-			if (n < 0L) return(-1);
-		}
-		*resultp = n;
-	}
-	else ptr = -1;
+	else return(-1);
 
 	while (s[ptr] && strchr(IFS_SET, s[ptr])) ptr++;
+
 	return(ptr);
 }
 
@@ -741,7 +813,7 @@ char *evalposixsubst(s, ptrp)
 char *s;
 int *ptrp;
 {
-	char *cp, buf[MAXLONGWIDTH + 1];
+	char *cp;
 	long n;
 	int i, len;
 
@@ -758,10 +830,10 @@ int *ptrp;
 		redirectlist red;
 
 		red.fd = red.type = red.new = 0;
-		red.filename = "";
+		red.filename = nullstr;
 		i = len = 0;
 		stree = newstree(NULL);
-		trp = startvar(stree, &red, "", &i, &len, 0);
+		trp = startvar(stree, &red, nullstr, &i, &len, 0);
 		trp -> cont = CN_COMM;
 		trp = analyze(s, trp, -1);
 		if (trp && (trp -> cont & CN_SBST) != CN_CASE) trp = NULL;
@@ -774,7 +846,7 @@ int *ptrp;
 	}
 	else {
 		n = 0L;
-		if (!(cp = evalvararg(s, 0, 1, '\'', 0, 0))) *ptrp = -1;
+		if (!(cp = evalvararg(s, '\'', EA_BACKQ, 0))) *ptrp = -1;
 		else {
 			for (i = 0; cp[i] && strchr(IFS_SET, cp[i]); i++);
 			if (cp[i]) {
@@ -785,14 +857,7 @@ int *ptrp;
 		}
 
 		if (*ptrp < 0) cp = NULL;
-		else if (n >= 0L)
-			cp = strdup2(long2str(buf, n, sizeof(buf)));
-		else {
-			long2str(buf, -n, sizeof(buf));
-			cp = malloc2(strlen(buf) + 2);
-			cp[0] = '-';
-			strcpy(&(cp[1]), buf);
-		}
+		else if (asprintf2(&cp, "%ld", n) < 0) error("malloc()");
 	}
 	free(s);
 
@@ -882,7 +947,7 @@ int reset;
 			free(mailpathlist[i].path);
 			if (mailpathlist[i].msg) free(mailpathlist[i].msg);
 		}
-		free(mailpathlist);
+		if (mailpathlist) free(mailpathlist);
 		mailpathlist = NULL;
 		maxmailpath = 0;
 		return;
@@ -908,7 +973,7 @@ syntaxtree *trp;
 	int i, j;
 
 	if (mypid != orgpgrp) return(RET_SUCCESS);
-	checkjob(0);
+	checkjob(NULL);
 	for (i = 0; i < maxjobs; i++) {
 		if (!(joblist[i].pids)) continue;
 		j = joblist[i].npipe;
@@ -928,6 +993,7 @@ syntaxtree *trp;
 			joblist[i].pids = NULL;
 		}
 	}
+
 	return(RET_SUCCESS);
 }
 
@@ -940,9 +1006,10 @@ syntaxtree *trp;
 	int i, j, n, ret;
 
 	s = ((trp -> comm) -> argc > 1) ? (trp -> comm) -> argv[1] : NULL;
-	checkjob(0);
+	checkjob(NULL);
 	if ((i = getjob(s)) < 0) {
-		execerror((trp -> comm) -> argv[1], ER_NOSUCHJOB, 0);
+		execerror((trp -> comm) -> argv[1],
+			(i < -1) ? ER_TERMINATED : ER_NOSUCHJOB, 0);
 		return(RET_FAIL);
 	}
 	n = joblist[i].npipe;
@@ -951,17 +1018,13 @@ syntaxtree *trp;
 		prepareexit(-1);
 		Xexit(RET_FATALERR);
 	}
-	fputc('[', stderr);
-	fputlong(i + 1, stderr);
-	fputs("] ", stderr);
-	fputlong(joblist[i].pids[n], stderr);
-	fputc('\n', stderr);
-	fflush(stderr);
+	fprintf2(stderr, "[%d] %id", i + 1, joblist[i].pids[n]);
+	fputnl(stderr);
 	if (joblist[i].tty) tioctl(ttyio, REQSETP, joblist[i].tty);
 # ifdef	USESGTTY
 	if (joblist[i].ttyflag) ioctl(ttyio, TIOCLSET, joblist[i].ttyflag);
 # endif
-	gettermio(joblist[i].pids[0]);
+	gettermio(joblist[i].pids[0], jobok);
 	if ((j = joblist[i].stats[n]) > 0 && j < 128) {
 		Xkillpg(joblist[i].pids[0], SIGCONT);
 		for (j = 0; j <= n; j++) joblist[i].stats[j] = 0;
@@ -969,12 +1032,13 @@ syntaxtree *trp;
 	if (isopbg(joblist[i].trp) || isopnown(joblist[i].trp))
 		joblist[i].trp -> type = OP_NONE;
 	ret = waitchild(joblist[i].pids[n], NULL);
-	gettermio(orgpgrp);
+	gettermio(orgpgrp, jobok);
 	if (tioctl(ttyio, REQSETP, &tty) < 0) {
 		doperror((trp -> comm) -> argv[0], "fatal error");
 		prepareexit(-1);
 		Xexit(RET_FATALERR);
 	}
+
 	return(ret);
 }
 
@@ -986,9 +1050,10 @@ syntaxtree *trp;
 	int i, j, n;
 
 	s = ((trp -> comm) -> argc > 1) ? (trp -> comm) -> argv[1] : NULL;
-	checkjob(0);
+	checkjob(NULL);
 	if ((i = getjob(s)) < 0) {
-		execerror((trp -> comm) -> argv[1], ER_NOSUCHJOB, 0);
+		execerror((trp -> comm) -> argv[1],
+			(i < -1) ? ER_TERMINATED : ER_NOSUCHJOB, 0);
 		return(RET_FAIL);
 	}
 	n = joblist[i].npipe;
@@ -1001,6 +1066,7 @@ syntaxtree *trp;
 	if (interactive && !nottyout) dispjob(i, stderr);
 	prevjob = lastjob;
 	lastjob = i;
+
 	return(RET_SUCCESS);
 }
 #endif	/* !NOJOB */
@@ -1019,14 +1085,11 @@ syntaxtree *trp;
 		for (i = 0; alias[i].ident; i++);
 		if (i > 1) qsort(alias, i, sizeof(shaliastable), cmpalias);
 		for (i = 0; alias[i].ident; i++) {
-			fputs("alias ", stdout);
-			kanjifputs(alias[i].ident, stdout);
-			fputs("='", stdout);
-			kanjifputs(alias[i].comm, stdout);
-			fputs("'\n", stdout);
+			fprintf2(stdout, "alias %k='%k'",
+				alias[i].ident, alias[i].comm);
+			fputnl(stdout);
 		}
 		freealias(alias);
-		fflush(stdout);
 		return(RET_SUCCESS);
 	}
 
@@ -1037,12 +1100,11 @@ syntaxtree *trp;
 			set = 1;
 			break;
 		}
-		for (i = 0; shellalias[i].ident; i++)
-			if (!strncommcmp(shellalias[i].ident, argv[n], len)
-			&& !(shellalias[i].ident[len]))
-				break;
 
-		if (!set) {
+		if (set) VOID_C addalias(strndup2(argv[n], len),
+			strdup2(&(argv[n][len + 1])));
+		else {
+			i = searchalias(argv[n], len);
 			if (!(shellalias[i].ident)) {
 				if (interactive) {
 					fputs("alias: ", stderr);
@@ -1051,58 +1113,25 @@ syntaxtree *trp;
 				ret = RET_FAIL;
 				ERRBREAK;
 			}
-			fputs("alias ", stdout);
-			kanjifputs(shellalias[i].ident, stdout);
-			fputs("='", stdout);
-			kanjifputs(shellalias[i].comm, stdout);
-			fputs("'\n", stdout);
-			fflush(stdout);
-		}
-		else if (shellalias[i].ident) {
-			free(shellalias[i].comm);
-			shellalias[i].comm = strdup2(&(argv[n][++len]));
-		}
-		else {
-			shellalias = (shaliastable *)realloc2(shellalias,
-				(i + 2) * sizeof(shaliastable));
-			shellalias[i].ident = strndup2(argv[n], len);
-			shellalias[i].comm = strdup2(&(argv[n][++len]));
-			shellalias[++i].ident = NULL;
+			fprintf2(stdout, "alias %k='%k'",
+				shellalias[i].ident, shellalias[i].comm);
+			fputnl(stdout);
 		}
 	}
+
 	return(ret);
 }
 
 int posixunalias(trp)
 syntaxtree *trp;
 {
-	reg_t *re;
 	char **argv;
-	int i, n, c, ret;
+	int n, ret;
 
 	argv = (trp -> comm) -> argv;
 	ret = RET_SUCCESS;
 	for (n = 1; n < (trp -> comm) -> argc; n++) {
-		c = 0;
-		re = regexp_init(argv[n], -1);
-		for (i = 0; shellalias[i].ident; i++) {
-			if (re) {
-				if (!regexp_exec(re, shellalias[i].ident, 0))
-					continue;
-			}
-			else if (strcommcmp(shellalias[i].ident, argv[n]))
-				continue;
-			c++;
-			free(shellalias[i].ident);
-			free(shellalias[i].comm);
-			for (; shellalias[i + 1].ident; i++) {
-				shellalias[i].ident = shellalias[i + 1].ident;
-				shellalias[i].comm = shellalias[i + 1].comm;
-			}
-			shellalias[i].ident = NULL;
-		}
-		if (re) regexp_free(re);
-		if (!c) {
+		if (deletealias(argv[n]) < 0) {
 			if (interactive) {
 				fputs("alias: ", stderr);
 				execerror(argv[n], ER_NOTALIAS, 0);
@@ -1111,6 +1140,7 @@ syntaxtree *trp;
 			ERRBREAK;
 		}
 	}
+
 	return(ret);
 }
 #endif	/* NOALIAS */
@@ -1124,8 +1154,8 @@ syntaxtree *trp;
 
 	if ((trp -> comm) -> argc <= 1) {
 		fputs("usage: kill [ -sig ] pid ...\n", stderr);
-		fputs("for a list of signals: kill -l\n", stderr);
-		fflush(stderr);
+		fputs("for a list of signals: kill -l", stderr);
+		fputnl(stderr);
 		return(RET_SYNTAXERR);
 	}
 #ifdef	SIGTERM
@@ -1141,10 +1171,11 @@ syntaxtree *trp;
 				for (i = 0; signallist[i].sig >= 0; i++)
 					if (sig == signallist[i].sig) break;
 				if (signallist[i].sig < 0) continue;
-				fputs(signallist[i].ident, stdout);
-				fputc((++n % 16) ? ' ' : '\n', stdout);
+				fprintf2(stdout, "%s%c",
+					signallist[i].ident,
+					(++n % 16) ? ' ' : '\n');
 			}
-			if (n % 16) fputc('\n', stdout);
+			if (n % 16) fputnl(stdout);
 			fflush(stdout);
 			return(RET_SUCCESS);
 		}
@@ -1166,24 +1197,25 @@ syntaxtree *trp;
 	}
 
 #ifndef	NOJOB
-	checkjob(0);
+	checkjob(NULL);
 #endif
 	for (; i < (trp -> comm) -> argc; i++) {
 		s = (trp -> comm) -> argv[i];
 #ifndef	NOJOB
 		if (*s == '%') {
 			if ((n = getjob(s)) < 0) {
-				execerror(s, ER_NOSUCHJOB, 0);
+				execerror(s, (n < -1)
+					? ER_TERMINATED : ER_NOSUCHJOB, 0);
 				return(RET_FAIL);
 			}
 			n = Xkillpg(joblist[n].pids[0], sig);
 		}
 		else
 #endif	/* !NOJOB */
-		if ((pid = isnumeric(s)) < 0L) {
+		if ((pid = isnumeric(s)) < (p_id_t)0) {
 			fputs("usage: kill [ -sig ] pid ...\n", stderr);
-			fputs("for a list of signals: kill -l\n", stderr);
-			fflush(stderr);
+			fputs("for a list of signals: kill -l", stderr);
+			fputnl(stderr);
 			return(RET_SYNTAXERR);
 		}
 #if	MSDOS
@@ -1199,6 +1231,7 @@ syntaxtree *trp;
 			return(RET_FAIL);
 		}
 	}
+
 	return(RET_SUCCESS);
 }
 
@@ -1264,6 +1297,14 @@ int *ptrp;
 			&& st.st_mode & S_ISVTX)
 				? RET_SUCCESS : RET_FAIL;
 			break;
+#if	defined (BASHSTYLE) || defined (STRICTPOSIX)
+	/* "test" on bash & POSIX has -e & -L option */
+		case 'e':
+			if (s) ret = (*s && Xlstat(s, &st) >= 0)
+				? RET_SUCCESS : RET_FAIL;
+			break;
+		case 'L':
+#endif
 		case 'h':
 			if (s) ret = (*s && Xlstat(s, &st) >= 0
 			&& (st.st_mode & S_IFMT) == S_IFLNK)
@@ -1301,18 +1342,12 @@ int *ptrp;
 		case 'n':
 			if (s) ret = (*s) ? RET_SUCCESS : RET_FAIL;
 			break;
-#ifdef	BASHSTYLE
-	/* bash's "test" has -e option */
-		case 'e':
-			if (s) ret = (*s && Xstat(s, &st) >= 0)
-				? RET_SUCCESS : RET_FAIL;
-			break;
-#endif
 		default:
 			ret = -2;
 			break;
 	}
 	if (ret >= 0) *ptrp += 2;
+
 	return(ret);
 }
 
@@ -1330,7 +1365,8 @@ int *ptrp;
 		if ((ret = testsub3(argc, argv, ptrp, 0)) < 0)
 			return(ret);
 		if (*ptrp >= argc
-		|| argv[*ptrp][0] != ')' || argv[*ptrp][1]) return(-1);
+		|| argv[*ptrp][0] != ')' || argv[*ptrp][1])
+			return(-1);
 		(*ptrp)++;
 		return(ret);
 	}
@@ -1391,6 +1427,7 @@ int *ptrp;
 	}
 	ret = (argv[*ptrp][0]) ? RET_SUCCESS : RET_FAIL;
 	(*ptrp)++;
+
 	return(ret);
 }
 
@@ -1422,6 +1459,7 @@ int *ptrp, lvl;
 		? (ret1 != RET_FAIL && ret2 != RET_FAIL)
 		: (ret1 != RET_FAIL || ret2 != RET_FAIL))
 			? RET_SUCCESS : RET_FAIL;
+
 	return(ret1);
 }
 
@@ -1437,8 +1475,8 @@ syntaxtree *trp;
 	if (argc <= 0) return(RET_FAIL);
 	if (argv[0][0] == '[' && !argv[0][1]
 	&& (--argc <= 0 || argv[argc][0] != ']' || argv[argc][1])) {
-		fputs("] missing\n", stderr);
-		fflush(stderr);
+		fputs("] missing", stderr);
+		fputnl(stderr);
 		return(RET_NOTICE);
 	}
 	if (argc <= 1) return(RET_FAIL);
@@ -1448,20 +1486,20 @@ syntaxtree *trp;
 	if (ret < 0) {
 		switch (ret) {
 			case -1:
-				fputs("argument expected\n", stderr);
+				fputs("argument expected", stderr);
 				break;
 			case -2:
-				fputs("unknown operator ", stderr);
-				fputs(argv[ptr], stderr);
-				fputc('\n', stderr);
+				fprintf2(stderr, "unknown operator %k",
+					argv[ptr]);
 				break;
 			default:
-				fputs("syntax error\n", stderr);
+				fputs("syntax error", stderr);
 				break;
 		}
-		fflush(stderr);
+		fputnl(stderr);
 		return(RET_NOTICE);
 	}
+
 	return(ret);
 }
 
@@ -1494,7 +1532,6 @@ syntaxtree *trp;
 # endif
 			}
 		}
-		fflush(stdout);
 		return(ret);
 	}
 	else if (flag == 'v') {
@@ -1521,14 +1558,13 @@ syntaxtree *trp;
 			}
 
 			if (cp) {
-				fputs(cp, stdout);
-				fputc('\n', stdout);
+				kanjifputs(cp, stdout);
+				fputnl(stdout);
 # ifdef	BASHSTYLE
 				ret = RET_SUCCESS;
 # endif
 			}
 		}
-		fflush(stdout);
 		return(ret);
 	}
 
@@ -1538,14 +1574,10 @@ syntaxtree *trp;
 	type = checktype(argv[n], &id, 0, 0);
 #endif
 	if (verboseexec) {
-		fputs("+ ", stderr);
-		kanjifputs(argv[n], stderr);
-		for (i = n + 1; i < argc; i++) {
-			fputc(' ', stderr);
-			kanjifputs(argv[i], stderr);
-		}
-		fputc('\n', stderr);
-		fflush(stderr);
+		fprintf2(stderr, "+ %k", argv[n]);
+		for (i = n + 1; i < argc; i++)
+			fprintf2(stderr, " %k", argv[i]);
+		fputnl(stderr);
 	}
 
 	(trp -> comm) -> argc -= n;
@@ -1557,18 +1589,19 @@ syntaxtree *trp;
 		ret = exec_simplecom(trp, type, id, 0);
 #endif
 	else {
-		path = strdup2(getconstvar("PATH"));
-		setenv2("PATH", DEFPATH, 1);
+		path = strdup2(getconstvar(ENVPATH));
+		setenv2(ENVPATH, DEFPATH, 1);
 #if	MSDOS
 		ret = exec_simplecom(trp, type, id);
 #else
 		ret = exec_simplecom(trp, type, id, 0);
 #endif
-		setenv2("PATH", path, 1);
+		setenv2(ENVPATH, path, 1);
 		if (path) free(path);
 	}
 	(trp -> comm) -> argc = argc;
 	(trp -> comm) -> argv = argv;
+
 	return(ret);
 }
 
@@ -1635,7 +1668,7 @@ syntaxtree *trp;
 	if (ret != RET_SUCCESS) n = '?';
 	else {
 		n = argv[posixoptind][ptr++];
-		if (!islower(n & 0xff) || !(cp = strchr(optstr, n))) {
+		if (n == ':' || ismsb(n) || !(cp = strchr2(optstr, n))) {
 			buf[0] = n;
 			buf[1] = '\0';
 			n = '?';
@@ -1663,15 +1696,16 @@ syntaxtree *trp;
 		}
 	}
 
-	if (posixoptarg) setenv2("OPTARG", posixoptarg, 0);
-	else if (ret == RET_SUCCESS) unset("OPTARG", sizeof("OPTARG") - 1);
+	if (posixoptarg) setenv2(ENVOPTARG, posixoptarg, 0);
+	else if (ret == RET_SUCCESS) unset(ENVOPTARG, strsize(ENVOPTARG));
 
 	buf[0] = n;
 	buf[1] = '\0';
 	setenv2(name, buf, 0);
 
 	n = posixoptind;
-	setenv2("OPTIND", long2str(buf, n, sizeof(buf)), 0);
+	snprintf2(buf, sizeof(buf), "%d", n);
+	setenv2(ENVOPTIND, buf, 0);
 	posixoptind = n;
 
 	return(ret);
