@@ -13,10 +13,6 @@
 #include "kanji.h"
 #include "termemu.h"
 
-#ifdef	USESELECTH
-#include <sys/select.h>
-#endif
-
 #ifdef	_NOORIGSHELL
 #include <signal.h>
 #include "termio.h"
@@ -29,18 +25,12 @@ extern int maxjobs;
 # endif
 #endif
 
-#ifndef	FD_SET
-typedef struct fd_set {
-	u_int fds_bits[1];
-} fd_set;
-# define	FD_ZERO(p)	(((p) -> fds_bits[0]) = 0)
-# define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= ((u_int)1 << (n)))
-#endif	/* !FD_SET */
-
-extern int internal_status;
+extern char fullpath[];
 extern int hideclock;
+extern int fdmode;
 
 int ptymode = 0;
+int ptyinternal = 0;
 char *ptyterm = NULL;
 int ptymenukey = -1;
 ptyinfo_t ptylist[MAXWINDOWS];
@@ -50,6 +40,7 @@ int parentfd = -1;
 char *ptytmpfile = NULL;
 
 static VOID NEAR doscroll __P_((int, int, int, int));
+static VOID NEAR closeallpty __P_((int));
 static int NEAR genbackend __P_((VOID_A));
 static VOID NEAR sendvar __P_((int, char **));
 #ifndef	_NOORIGSHELL
@@ -93,7 +84,7 @@ int n, c, x, y, min, max;
 	}
 
 	y1 = y2 = -1;
-	switch(n) {
+	switch (n) {
 		case C_DOWN:
 		case C_SCROLLFORW:
 		case C_NEWLINE:
@@ -131,29 +122,12 @@ int n, c, x, y, min, max;
 	}
 }
 
-int selectpty(fd, fds, result, timeout)
-int fd, fds[];
+int selectpty(max, fds, result, timeout)
+int max, fds[];
 char result[];
 int timeout;
 {
-	fd_set readfds;
 	struct timeval tv, *t;
-	int i, n, max;
-
-	max = -1;
-	FD_ZERO(&readfds);
-	if (fd < 0) max = 0;
-	else {
-		max = fd;
-		FD_SET(fd, &readfds);
-	}
-	for (i = 0; i < MAXWINDOWS; i++) {
-		if (!(ptylist[i].pid)) continue;
-		if (fds[i] > max) max = fds[i];
-		FD_SET(fds[i], &readfds);
-	}
-
-	if (!max) return(0);
 
 	if (timeout < 0) t = NULL;
 	else if (!timeout) {
@@ -167,13 +141,18 @@ int timeout;
 		t = &tv;
 	}
 
-	if ((n = select(max + 1, &readfds, NULL, NULL, t)) < 0) return(-1);
-	for (i = 0; i < MAXWINDOWS; i++)
-		result[i] = (ptylist[i].pid && FD_ISSET(fds[i], &readfds))
-			? 1 : 0;
-	if (fd >= 0) result[i] = (FD_ISSET(fd, &readfds)) ? 1 : 0;
+	return(readselect(max, fds, result, t));
+}
 
-	return(n);
+static VOID NEAR closeallpty(n)
+int n;
+{
+	int i;
+
+	for (i = 0; i < n; i++) {
+		safeclose(ptylist[i].fd);
+		ptylist[i].fd = -1;
+	}
 }
 
 static int NEAR genbackend(VOID_A)
@@ -199,11 +178,10 @@ static int NEAR genbackend(VOID_A)
 	if (i < MAXWINDOWS || (pid = Xfork()) < (p_id_t)0) {
 		safeclose(fds[0]);
 		safeclose(fds[1]);
+		closeallpty(i);
 		while (--i >= 0) {
 			free(ptylist[i].path);
 			ptylist[i].path = NULL;
-			safeclose(ptylist[i].fd);
-			ptylist[i].fd = -1;
 		}
 		return(-1);
 	}
@@ -223,17 +201,18 @@ static int NEAR genbackend(VOID_A)
 		resetptyterm(i, 1);
 
 		backend();
-		for (i = 0; i < MAXWINDOWS; i++) safeclose(ptylist[i].fd);
+		closeallpty(MAXWINDOWS);
 		_exit(0);
 	}
 
 	safeclose(fds[0]);
-	emufd = fds[1];
+	emufd = newdup(fds[1]);
 	emupid = pid;
-	for (i = 0; i < MAXWINDOWS; i++) {
-		safeclose(ptylist[i].fd);
-		ptylist[i].fd = -1;
-	}
+#ifdef	CYGWIN
+	for (i = 0; i < MAXWINDOWS; i++) ptylist[i].fd = newdup(ptylist[i].fd);
+#else
+	closeallpty(MAXWINDOWS);
+#endif
 
 	return(0);
 }
@@ -568,6 +547,9 @@ va_dcl
 #endif
 #if	!defined (_NOORIGSHELL) && !defined (NOJOB)
 		case TE_CHANGESTATUS:
+#endif
+#ifndef	_NOIME
+		case TE_FREEROMAN:
 #endif
 			n = va_arg(args, int);
 			sendbuf(fd, &n, sizeof(n));
@@ -609,6 +591,27 @@ va_dcl
 			sendstring(fd, devp -> name);
 			break;
 #endif	/* _USEDOSEMU */
+		case TE_SAVETTYIO:
+			n = va_arg(args, int);
+			val = (n >= 0 && duptty[n]) ? TIO_BUFSIZ : 0;
+			sendbuf(fd, &n, sizeof(n));
+			sendbuf(fd, &val, sizeof(val));
+			if (n >= 0 && duptty[n]) sendbuf(fd, duptty[n], val);
+			break;
+#ifndef	_NOIME
+		case TE_ADDROMAN:
+			cp = va_arg(args, char *);
+			func1 = va_arg(args, char *);
+			sendstring(fd, cp);
+			sendstring(fd, func1);
+			break;
+#endif
+		case TE_INTERNAL:
+			n = va_arg(args, int);
+			cp = va_arg(args, char *);
+			sendbuf(fd, &n, sizeof(n));
+			sendstring(fd, cp);
+			break;
 		default:
 			break;
 	}
@@ -630,6 +633,7 @@ int flags;
 	sendbuf(emufd, &flags, sizeof(flags));
 	sendstring(emufd, command);
 	sendstring(emufd, arg);
+	sendstring(emufd, fullpath);
 }
 
 #if	!defined (_NOORIGSHELL) && !defined (NOJOB)
@@ -656,7 +660,7 @@ static int NEAR recvmacro(commandp, argp, flagsp)
 char **commandp, **argp;
 int *flagsp;
 {
-	char *command, *arg;
+	char *command, *arg, *cwd;
 	int flags;
 
 	Xttyiomode(1);
@@ -667,6 +671,10 @@ int *flagsp;
 		if (command) free(command);
 		return(-1);
 	}
+	if (recvstring(ttyio, &cwd) >= 0 && cwd) {
+		VOID_C chdir2(cwd);
+		free(cwd);
+	}
 
 	keyflush();
 	if (!(flags & F_TTYIOMODE)) Xstdiomode();
@@ -689,16 +697,18 @@ int flags;
 	int i, n;
 
 	if (parentfd < 0) {
+		changewin(MAXWINDOWS, (p_id_t)-1);
 		if (ptymode) {
-			hideclock = 1;
+			hideclock = 2;
 			warning(0, NOPTY_K);
 		}
+		killpty(win, NULL);
 		for (i = 0; i < MAXWINDOWS; i++) if (ptylist[i].pid) break;
 		if (i < MAXWINDOWS) {
-			hideclock = 1;
+			hideclock = 2;
 			if (!yesno(KILL_K)) return(-1);
-			killallpty();
 		}
+		killallpty();
 	}
 
 	if (flags & F_DOSYSTEM) n = dosystem(command);
@@ -714,12 +724,22 @@ int ptymacro(command, arg, flags)
 char *command, *arg;
 int flags;
 {
+#if	!defined (_NOORIGSHELL) && !defined (NOJOB)
+	char *command2, *arg2;
+#endif
 	p_id_t pid;
 	char *tty, *ws, path[MAXPATHLEN], buf[TIO_BUFSIZ + TIO_WINSIZ];
 	u_char uc;
 	int i, n, fd, fds[2];
 
-	if (!ptymode) return(callmacro(command, arg, flags));
+#ifndef	_NOORIGSHELL
+	if (isshptymode()) {
+		if (!(flags & F_DOSYSTEM))
+			return(callmacro(command, arg, flags));
+	}
+	else
+#endif
+	if (!ptymode || ptyinternal) return(callmacro(command, arg, flags));
 
 	if (ptylist[win].pid && emufd >= 0) {
 		awakechild(command, arg, flags);
@@ -749,6 +769,11 @@ int flags;
 		VOID_C Xclose(fd);
 	}
 
+#ifndef	_NOKANJICONV
+	changeinkcode();
+	changeoutkcode();
+#endif
+
 	n = sigvecset(0);
 	if ((pid = Xfork()) < (p_id_t)0) {
 		if (fd >= 0 && ptytmpfile) {
@@ -765,8 +790,15 @@ int flags;
 	if (!pid) {
 #ifndef	_NOORIGSHELL
 		mypid = getpid();
+#endif
+#ifdef	CYGWIN
+		closeallpty(MAXWINDOWS);
 #endif
 		if (Xlogin_tty(ptylist[win].path, tty, ws) < 0) _exit(1);
+#ifndef	_NOORIGSHELL
+		if (isshptymode()) /*EMPTY*/;
+		else
+#endif
 		n_line = FILEPERROW;
 		VOID_C setwsize(STDIN_FILENO, n_column, n_line);
 
@@ -788,7 +820,6 @@ int flags;
 			dup2(fd, ttyio);
 			safeclose(fd);
 		}
-		maxfile = -1;
 
 		setdefterment();
 		setdefkeyseq();
@@ -799,6 +830,9 @@ int flags;
 		uc = '\n';
 		sendbuf(parentfd, &uc, sizeof(uc));
 
+#if	!defined (_NOORIGSHELL) && !defined (NOJOB)
+		command2 = arg2 = NULL;
+#endif
 		for (;;) {
 			syncptyout(-1, -1);
 			setlinecol();
@@ -811,7 +845,9 @@ int flags;
 			if (i >= maxjobs) break;
 			VOID_C signal2(SIGHUP, (sigcst_t)trap_hup);
 			sendparent(TE_CHANGESTATUS, n);
-			if (recvmacro(&command, &arg, &flags) < 0) break;
+			if (recvmacro(&command2, &arg2, &flags) < 0) break;
+			command = command2;
+			arg = arg2;
 #endif	/* !_NOORIGSHELL && !NOJOB */
 		}
 
@@ -822,7 +858,7 @@ int flags;
 
 	safeclose(fds[1]);
 	ptylist[win].pid = pid;
-	ptylist[win].pipe = fds[0];
+	ptylist[win].pipe = newdup(fds[0]);
 	ptylist[win].status = -1;
 
 	sigvecset(n);
@@ -843,8 +879,9 @@ int n, *statusp;
 		VOID_C waitstatus(ptylist[n].pid, 0, statusp);
 		ptylist[n].pid = (p_id_t)0;
 		ptylist[n].status = 0;
-		changewin(n, (p_id_t)0);
 	}
+	changewin(n, (p_id_t)0);
+	changewin(MAXWINDOWS, (p_id_t)-1);
 
 	safeclose(ptylist[n].pipe);
 	ptylist[n].pipe = -1;
@@ -856,10 +893,15 @@ VOID killallpty(VOID_A)
 
 	for (i = 0; i < MAXWINDOWS; i++) {
 		killpty(i, NULL);
-		free(ptylist[i].path);
-		ptylist[i].path = NULL;
+		if (ptylist[i].path) {
+			free(ptylist[i].path);
+			ptylist[i].path = NULL;
+		}
 	}
 
+#ifdef	CYGWIN
+	closeallpty(MAXWINDOWS);
+#endif
 	if (emupid) {
 		VOID_C kill(emupid, SIGHUP);
 #ifdef	SIGCONT
@@ -877,4 +919,28 @@ VOID killallpty(VOID_A)
 	safeclose(emufd);
 	emufd = -1;
 }
+
+int checkpty(n)
+int n;
+{
+	int status;
+
+	if (!(ptylist[n].pid)) return(0);
+	if (waitstatus(ptylist[n].pid, WNOHANG, &status) < 0) return(0);
+	ptylist[n].pid = (p_id_t)0;
+	ptylist[n].status = status;
+	killpty(n, &status);
+
+	return(-1);
+}
+
+int checkallpty(VOID_A)
+{
+	int i, n;
+
+	n = 0;
+	for (i = 0; i < MAXWINDOWS; i++) if (checkpty(i) < 0) n = -1;
+
+	return(n);
+}
 #endif	/* !_NOPTY */
