@@ -4,10 +4,10 @@
  *	UNIX-like disk accessing module
  */
 
+#include "machine.h"
 #include <stdio.h>
 #include <fcntl.h>
 #include <io.h>
-#include "machine.h"
 
 #ifndef	NOUNISTDH
 #include <unistd.h>
@@ -38,21 +38,29 @@
 
 #ifdef	FD
 extern int _dospath __P_((char *));
+extern char *gendospath __P_((char *, int, int));
 extern char *strdelim __P_((char *, int));
 extern char *strrdelim __P_((char *, int));
 extern int isdelim __P_((char *, int));
 extern char *strcatdelim __P_((char *));
 extern char *strcatdelim2 __P_((char *, char *, char *));
 extern int isdotdir __P_((char *));
+# ifdef	DOUBLESLASH
+extern int isdslash __P_((char *));
+# endif
 extern char curpath[];
 #else
 static int NEAR _dospath __P_((char *));
+static char *NEAR gendospath __P_((char *, int, int));
 static char *NEAR strdelim __P_((char *, int));
 static char *NEAR strrdelim __P_((char *, int));
 static int NEAR isdelim __P_((char *, int));
 static char *NEAR strcatdelim __P_((char *));
 static char *NEAR strcatdelim2 __P_((char *, char *, char *));
 static int NEAR isdotdir __P_((char *));
+# ifdef	DOUBLESLASH
+static int isdslash __P_((char *));
+# endif
 static char curpath[] = ".";
 #endif
 
@@ -94,6 +102,8 @@ static time_t NEAR getdostime __P_((u_int, u_int));
 #ifndef	_NOUSELFN
 static int NEAR putdostime __P_((u_short *, u_short *, time_t));
 #endif
+static int NEAR alter_findfirst __P_((char *, u_int *,
+		struct dosfind_t *, struct lfnfind_t *));
 
 #ifndef	_NODOSDRIVE
 int dosdrive = 0;
@@ -142,6 +152,9 @@ static u_char int25bin[] = {
 };
 static int (far *doint25)__P_((VOID_A)) = NULL;
 #endif	/* !DJGPP && !_NODOSDRIVE */
+#ifdef	DOUBLESLASH
+static int dsdrive = 0;
+#endif
 
 
 #ifndef	FD
@@ -151,6 +164,18 @@ char *path;
 	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
 }
 
+static char *NEAR gendospath(path, drive, c)
+char *path;
+int drive, c;
+{
+	*(path++) = drive;
+	*(path++) = ':';
+	if (c) *(path++) = c;
+	*path = '\0';
+
+	return(path);
+}
+
 static char *NEAR strdelim(s, d)
 char *s;
 int d;
@@ -302,6 +327,19 @@ char *s;
 
 	return(0);
 }
+
+# ifdef	DOUBLESLASH
+static int isdslash(s)
+char *s;
+{
+	char *cp;
+
+	if (s[0] != _SC_ || s[1] != _SC_ || !s[2] || s[2] == _SC_) return(0);
+	if ((cp = strdelim(&(s[2]), 0))) return(cp - s);
+
+	return(strlen(s));
+}
+# endif	/* DOUBLESLASH */
 #endif	/* !FD */
 
 #ifdef	DJGPP
@@ -321,6 +359,9 @@ int offset;
 static char *NEAR duplpath(path)
 char *path;
 {
+#ifdef	DOUBLESLASH
+	int len;
+#endif
 	static char buf[MAXPATHLEN];
 	int i, j, ps;
 
@@ -330,6 +371,13 @@ char *path;
 		buf[j++] = path[i++];
 		buf[j++] = path[i++];
 	}
+#ifdef	DOUBLESLASH
+	if ((len = isdslash(path))) {
+		memcpy(buf, path, len);
+		i = j = len;
+	}
+	else
+#endif
 	if (path[i] == _SC_) buf[j++] = path[i++];
 	for (; path[i]; i++) {
 		if (path[i] == _SC_) ps = 1;
@@ -351,6 +399,9 @@ int getcurdrv(VOID_A)
 {
 	int drive;
 
+#ifdef	DOUBLESLASH
+	if (dsdrive) return(dsdrive);
+#endif
 	drive = (dos7access & D7_CAPITAL) ? 'A' : 'a';
 #ifndef	_NODOSDRIVE
 	if (lastdrive >= 0) drive += toupper2(lastdrive) - 'A';
@@ -364,11 +415,22 @@ int getcurdrv(VOID_A)
 int setcurdrv(drive, nodir)
 int drive, nodir;
 {
+#ifndef	_NODOSDRIVE
+	char tmp[3];
+#endif
 	struct SREGS sreg;
 	__dpmi_regs reg;
 	char *path;
 	int drv, olddrv;
 
+#ifdef	DOUBLESLASH
+	if (drive == '_') {
+		dsdrive = drive;
+		return(0);
+	}
+	dsdrive = '\0';
+#endif
+
 	errno = EINVAL;
 	if (islower2(drive)) {
 		drv = drive - 'a';
@@ -382,11 +444,7 @@ int drive, nodir;
 
 #ifndef	_NODOSDRIVE
 	if (checkdrive(drv) > 0) {
-		char tmp[3];
-
-		tmp[0] = drive;
-		tmp[1] = ':';
-		tmp[2] = '\0';
+		VOID_C gendospath(tmp, drive, '\0');
 		return(doschdir(tmp));
 	}
 #endif
@@ -447,6 +505,7 @@ int getdosver(VOID_A)
 }
 
 /*
+ *	 3: enable LFN access, but shared folder
  *	 2: enable LFN access, but FAT32
  *	 1: enable LFN access
  *	 0: do SFN access
@@ -461,24 +520,27 @@ char *path;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
-	char drv[4], buf[128];
+	char drive, drv[4], buf[128];
 
-	if (!path || !(drv[0] = _dospath(path))) drv[0] = getcurdrv();
+#ifdef	DOUBLESLASH
+	if (isdslash(path)) return(3);
+#endif
+	if (!path || !(drive = _dospath(path))) drive = getcurdrv();
+#ifdef	DOUBLESLASH
+	if (drive == '_') return(3);
+#endif
 #ifndef	_NODOSDRIVE
-	if (checkdrive(toupper2(drv[0]) - 'A') > 0) return(-2);
+	if (checkdrive(toupper2(drive) - 'A') > 0) return(-2);
 #endif
 	if (getdosver() < 7) {
 #ifndef	_NODOSDRIVE
-		if ((dosdrive & 1) && islower2(drv[0])) return(-1);
+		if ((dosdrive & 1) && islower2(drive)) return(-1);
 #endif
 		return(-3);
 	}
-	if (isupper2(drv[0])) return(0);
-
-	drv[1] = ':';
-	drv[2] = '\\';
-	drv[3] = '\0';
+	if (isupper2(drive)) return(0);
 
+	VOID_C gendospath(drv, drive, _SC_);
 	reg.x.ax = 0x71a0;
 	reg.x.bx = 0;
 	reg.x.cx = sizeof(buf);
@@ -540,16 +602,9 @@ char *path, *buf;
 	int drive;
 
 	cp = path;
-	if ((drive = _dospath(path))) cp += 2;
-	else drive = getcurdrv();
-	buf[0] = drive;
-	buf[1] = ':';
-	if (!*cp) {
-		buf[2] = '.';
-		buf[3] = '\0';
-	}
-	else if (cp == &(path[2])) return(path);
-	else strcpy(&(buf[2]), cp);
+	if (!(drive = _dospath(path))) drive = getcurdrv();
+	else if (*(cp += 2)) return(path);
+	strcpy(gendospath(buf, drive, (*cp) ? '\0' : '.'), cp);
 
 	return(buf);
 }
@@ -592,6 +647,9 @@ char *path, *alias;
 # ifndef	BSPATHDELIM
 	adjustpname(alias);
 # endif
+# ifdef	DOUBLESLASH
+	if (isdslash(alias)) return(alias);
+# endif
 
 	if (!path || !(i = _dospath(path))) i = getcurdrv();
 	if (islower2(i)) *alias = tolower2(*alias);
@@ -616,6 +674,7 @@ char *path, *resolved;
 	reg.x.cx = 0;
 #else	/* !_NOUSELFN */
 	switch (supportLFN(path)) {
+		case 3:
 		case 2:
 		case 1:
 			reg.x.ax = 0x7160;
@@ -657,6 +716,9 @@ char *path, *resolved;
 #ifndef	BSPATHDELIM
 	adjustpname(resolved);
 #endif
+#ifdef	DOUBLESLASH
+	if (isdslash(resolved)) return(resolved);
+#endif
 
 	if (!path || !(i = _dospath(path))) i = getcurdrv();
 	else path += 2;
@@ -675,10 +737,7 @@ char *path, *resolved;
 			resolved[j + 3] = '\0';
 		}
 		else {
-			resolved[0] = i;
-			resolved[1] = ':';
-			resolved[2] = _SC_;
-			resolved[3] = '\0';
+			VOID_C gendospath(resolved, i, _SC_);
 			if (path && *path == _SC_)
 				strcpy(&(resolved[2]), path);
 		}
@@ -1552,7 +1611,10 @@ struct dosfind_t *result;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, sizeof(struct dosfind_t));
-	if (int21call(&reg, &sreg) < 0) return(-1);
+	if (int21call(&reg, &sreg) < 0) {
+		if (!errno) errno = ENOENT;
+		return(-1);
+	}
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct dosfind_t), result);
 #endif
@@ -1604,7 +1666,10 @@ struct lfnfind_t *result;
 	reg.x.dx = PTR_OFF(path, sizeof(struct lfnfind_t));
 	sreg.es = PTR_SEG(result);
 	reg.x.di = PTR_OFF(result, 0);
-	if (int21call(&reg, &sreg) < 0) return(-1);
+	if (int21call(&reg, &sreg) < 0) {
+		if (!errno) errno = ENOENT;
+		return(-1);
+	}
 	*fdp = reg.x.ax;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct lfnfind_t), result);
@@ -2132,6 +2197,7 @@ int size;
 #ifdef	DJGPP
 	char tmp[MAXPATHLEN];
 #endif
+	char *cp;
 	int drive;
 
 	drive = getcurdrv();
@@ -2143,11 +2209,12 @@ int size;
 #endif
 	if (!pathname && !(pathname = (char *)malloc(size))) return(NULL);
 	else {
-		pathname[0] = drive;
-		pathname[1] = ':';
-		pathname[2] = _SC_;
-
-		if (!unixgetcurdir(&(pathname[3]), 0)) return(NULL);
+#ifdef	DOUBLESLASH
+		if (drive == '_') cp = pathname;
+		else
+#endif
+		cp = gendospath(pathname, drive, _SC_);
+		if (!unixgetcurdir(cp, 0)) return(NULL);
 	}
 
 	*pathname = (dos7access & D7_CAPITAL)
@@ -2261,7 +2328,7 @@ statfs_t *buf;
 	}
 	else
 # endif
-	if (i == 2) {
+	if (i >= 2) {
 		reg.x.ax = 0x7303;
 		reg.x.cx = sizeof(fsbuf);
 # ifdef	DJGPP
@@ -2309,58 +2376,83 @@ statfs_t *buf;
 	return(0);
 }
 
+/*ARGSUSED*/
+static int NEAR alter_findfirst(path, fdp, dbuf, lbuf)
+char *path;
+u_int *fdp;
+struct dosfind_t *dbuf;
+struct lfnfind_t *lbuf;
+{
+	char *cp, tmp[MAXPATHLEN];
+	int n, len;
+
+#ifndef	_NOUSELFN
+	if (fdp) n = lfn_findfirst(fdp, path, SEARCHATTRS, lbuf);
+	else
+#endif
+	n = dos_findfirst(path, SEARCHATTRS, dbuf);
+	if (n >= 0) return(n);
+	if (errno != ENOENT) return(-1);
+
+	if ((cp = strrdelim(path, 1))) cp++;
+	else cp = path;
+
+	if (isdotdir(cp) == 1) {
+		len = cp - path;
+		memcpy(tmp, path, len);
+		strcpy(&(tmp[len]), curpath);
+#ifndef	_NOUSELFN
+		if (fdp) n = lfn_findfirst(fdp, tmp, SEARCHATTRS, lbuf);
+		else
+#endif
+		n = dos_findfirst(tmp, SEARCHATTRS, dbuf);
+	}
+#ifdef	DOUBLESLASH
+	else {
+		if (isdslash(path)) return(-1);
+		else if (_dospath(path)) /*EMPTY*/;
+		else if (getcurdrv() == '_') return(-1);
+#endif
+		n = dos_findfirst(path, DS_IFLABEL, dbuf);
+	}
+
+	return(n);
+}
+
 int unixstat(path, stp)
 char *path;
 struct stat *stp;
 {
 #ifndef	_NOUSELFN
+# ifndef	_NODOSDRIVE
+	char buf[MAXPATHLEN];
+# endif
 	struct lfnfind_t lbuf;
 	u_int fd;
-#endif
+#endif	/* !_NOUSELFN */
 	struct dosfind_t dbuf;
 	int i;
 
-#ifdef	_NOUSELFN
-	if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-		if (errno && errno != ENOENT) return(-1);
-		if (isdotdir(path) == 1)
-			i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
-		else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
+	if (strpbrk(path, "*?")) {
+		errno = ENOENT;
+		return(-1);
 	}
+#ifdef	_NOUSELFN
+	i = alter_findfirst(path, NULL, &dbuf, NULL);
 #else	/* !_NOUSELFN */
+	fd = (u_int)-1;
 	i = supportLFN(path);
 # ifndef	_NODOSDRIVE
 	if (i < 0 && i > -3) {
-		char buf[MAXPATHLEN];
-
 		dependdosfunc = 0;
 		if (dosstat(regpath(path, buf), stp) >= 0) return(0);
 		else if (!dependdosfunc) return(-1);
 	}
-# endif	/* !_NODOSDRIVE */
-	if (i <= 0) {
-		fd = (u_int)-1;
-		if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-			if (errno && errno != ENOENT) return(-1);
-			if (isdotdir(path) == 1)
-				i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
-			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
-		}
-	}
-	else {
-		if ((i = lfn_findfirst(&fd, path, SEARCHATTRS, &lbuf)) < 0) {
-			if (errno && errno != ENOENT) return(-1);
-			if (isdotdir(path) == 1)
-				i = lfn_findfirst(&fd, curpath,
-					SEARCHATTRS, &lbuf);
-			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
-		}
-	}
+# endif
+	if (i <= 0) i = alter_findfirst(path, NULL, &dbuf, &lbuf);
+	else i = alter_findfirst(path, &fd, &dbuf, &lbuf);
 #endif	/* !_NOUSELFN */
-	if (i < 0) {
-		if (!errno) errno = ENOENT;
-		return(-1);
-	}
+	if (i < 0) return(-1);
 
 #ifndef	_NOUSELFN
 	if (fd != (u_int)-1) {
@@ -2467,13 +2559,15 @@ struct timeval tvp[2];
 	}
 #endif	/* !USEUTIME */
 	if (i <= 0) {
-		if ((fd = open(path, O_RDONLY, 0666)) < 0) return(-1);
-		reg.x.ax = 0x5701;
-		reg.x.bx = (u_short)fd;
-		putdostime(&(reg.x.dx), &(reg.x.cx), t);
-		i = int21call(&reg, &sreg);
-		close(fd);
-		return(i);
+		if ((fd = open(path, O_RDONLY, 0666)) >= 0) {
+			reg.x.ax = 0x5701;
+			reg.x.bx = (u_short)fd;
+			putdostime(&(reg.x.dx), &(reg.x.cx), t);
+			i = int21call(&reg, &sreg);
+			close(fd);
+			return(i);
+		}
+		if (i || errno != EACCES) return(-1);
 	}
 
 	reg.x.ax = 0x7143;
@@ -2544,7 +2638,7 @@ FILE *unixfopen(path, type)
 char *path, *type;
 {
 	FILE *fp;
-	int i, flags;
+	int i, fd, flags;
 
 	path = duplpath(path);
 	i = supportLFN(path);
@@ -2583,8 +2677,8 @@ char *path, *type;
 			break;
 	}
 
-	if ((i = unixopen(path, flags, 0666)) < 0) return(NULL);
+	if ((fd = unixopen(path, flags, 0666)) < 0) return(NULL);
 
-	return(fdopen(i, type));
+	return(fdopen(fd, type));
 }
 #endif	/* !_NOUSELFN */
