@@ -10,6 +10,10 @@
 #include "func.h"
 #include "unixdisk.h"
 
+#ifndef	O_ACCMODE
+#define	O_ACCMODE	(O_RDONLY | O_WRONLY | O_RDWR)
+#endif
+
 #ifndef	_NOROCKRIDGE
 typedef struct _opendirpath_t {
 	DIR *dirp;
@@ -50,11 +54,15 @@ char *path, *buf;
 	}
 	if (buf) {
 #ifndef	_NOUSELFN
-		if (shortname(path, buf) == buf);
+		if (shortname(path, buf) == buf) /*EMPTY*/;
 		else
 #endif
-		strcpy(buf, path);
+		unixrealpath(path, buf);
 	}
+#ifdef	DOUBLESLASH
+	if (isdslash(path)) drv = '_';
+	else
+#endif
 	drv = _dospath(path);
 
 	return((drv) ? drv : getcurdrv());
@@ -76,9 +84,9 @@ char *path;
 int dospath3(path)
 char *path;
 {
-	int drive;
+	int i, drive;
 
-	if ((drive = supportLFN(path)) >= 0 || drive <= -3) return(0);
+	if ((i = supportLFN(path)) >= 0 || i <= -3) return(0);
 
 	return((drive = _dospath(path)) ? drive : getcurdrv());
 }
@@ -89,6 +97,9 @@ char *path, *buf;
 	char *cp, tmp[MAXPATHLEN];
 	int i, drive;
 
+#ifdef	DOUBLESLASH
+	if (isdslash(path)) return(0);
+#endif
 	if ((drive = _dospath(path))) cp = path + 2;
 	else {
 		cp = path;
@@ -102,10 +113,7 @@ char *path, *buf;
 		strcpy(tmp, cp);
 		cp = tmp;
 	}
-	if (*cp == _SC_) {
-		*(buf++) = drive;
-		*(buf++) = ':';
-	}
+	if (*cp == _SC_) buf = gendospath(buf, drive, '\0');
 	else {
 		if (!dosgetcwd(buf, MAXPATHLEN)) return(0);
 		buf = strcatdelim(buf);
@@ -249,8 +257,10 @@ char *path;
 			else dd = lastdrv;
 		}
 		lastdrv = dd;
+		n = 0;
 	}
 #endif	/* !_NODOSDRIVE */
+	LOG1(_LOG_INFO_, n, "chdir(\"%k\");", path);
 
 	return(n);
 }
@@ -280,9 +290,9 @@ struct stat *stp;
 	mode = (u_short)(stp -> st_mode);
 	if ((mode & S_IFMT) != S_IFDIR
 	&& (cp = strrchr(path, '.')) && strlen(++cp) == 3) {
-		if (!stricmp(cp, "BAT")
-		|| !stricmp(cp, "COM")
-		|| !stricmp(cp, "EXE"))
+		if (!stricmp(cp, EXTCOM)
+		|| !stricmp(cp, EXTEXE)
+		|| !stricmp(cp, EXTBAT))
 			mode |= S_IEXEC;
 	}
 	mode &= (S_IREAD | S_IWRITE | S_IEXEC);
@@ -363,6 +373,7 @@ int Xsymlink(name1, name2)
 char *name1, *name2;
 {
 	errno = EINVAL;
+	LOG2(_LOG_WARNING_, -1, "symlink(\"%k\", \"%k\");", name1, name2);
 
 	return(-1);
 }
@@ -397,6 +408,7 @@ int mode;
 
 	path = convput(conv, path, 1, 1, NULL, NULL);
 	n = unixchmod(path, mode);
+	LOG2(_LOG_NOTICE_, n, "chmod(\"%k\", %d);", path, mode);
 
 	return(n);
 }
@@ -411,6 +423,7 @@ struct utimbuf *times;
 
 	path = convput(conv, path, 1, 1, NULL, NULL);
 	n = unixutime(path, times);
+	LOG1(_LOG_NOTICE_, n, "utime(\"%k\");", path);
 
 	return(n);
 }
@@ -424,6 +437,7 @@ struct timeval tvp[2];
 
 	path = convput(conv, path, 1, 1, NULL, NULL);
 	n = unixutimes(path, tvp);
+	LOG1(_LOG_NOTICE_, n, "utimes(\"%k\");", path);
 
 	return(n);
 }
@@ -436,6 +450,7 @@ char *path;
 u_long flags;
 {
 	errno = EACCESS;
+	LOG2(_LOG_WARNING_, -1, "chflags(\"%k\", %05o);", path, flags);
 
 	return(-1);
 }
@@ -449,6 +464,7 @@ uid_t uid;
 gid_t gid;
 {
 	errno = EACCESS;
+	LOG3(_LOG_WARNING_, -1, "chown(\"%k\", %d, %d);", path, uid, gid);
 
 	return(-1);
 }
@@ -466,6 +482,7 @@ char *path;
 		&& unixchmod(path, (S_IREAD | S_IWRITE | S_ISVTX)) >= 0)
 			n = unixunlink(path);
 	}
+	LOG1(_LOG_WARNING_, n, "unlink(\"%k\");", path);
 
 	return(n);
 }
@@ -483,6 +500,7 @@ char *from, *to;
 		n = -1;
 	}
 	else n = unixrename(from, to);
+	LOG2(_LOG_WARNING_, n, "rename(\"%k\", \"%k\");", from, to);
 
 	return(n);
 }
@@ -508,6 +526,22 @@ int flags, mode;
 	else
 #endif	/* !_NOUSELFN */
 	fd = open(path, flags, mode);
+#ifndef	_NOLOGGING
+	switch (flags & O_ACCMODE) {
+		case O_WRONLY:
+			LOG2(_LOG_WARNING_, fd,
+				"open(\"%k\", O_WRONLY, %05o);", path, mode);
+			break;
+		case O_RDWR:
+			LOG2(_LOG_WARNING_, fd,
+				"open(\"%k\", O_RDWR, %05o);", path, mode);
+			break;
+		default:
+			LOG2(_LOG_INFO_, fd,
+				"open(\"%k\", O_RDONLY, %05o);", path, mode);
+			break;
+	}
+#endif	/* !_NOLOGGING */
 
 	return(fd);
 }
@@ -563,6 +597,18 @@ int whence;
 	return(ofs);
 }
 
+int Xftruncate(fd, len)
+int fd;
+off_t len;
+{
+	int n;
+
+	if (fd >= DOSFDOFFSET) n = dosftruncate(fd, len);
+	else n = ftruncate(fd, len);
+
+	return(n);
+}
+
 int Xdup(oldd)
 int oldd;
 {
@@ -614,6 +660,7 @@ int mode;
 		path = convput(conv, path, 1, 1, NULL, NULL);
 		n = unixmkdir(path, mode);
 	}
+	LOG2(_LOG_WARNING_, n, "mkdir(\"%k\", %05o);", path, mode);
 
 	return(n);
 }
@@ -626,6 +673,7 @@ char *path;
 
 	path = convput(conv, path, 1, 1, NULL, NULL);
 	n = unixrmdir(path);
+	LOG1(_LOG_WARNING_, n, "rmdir(\"%k\");", path);
 
 	return(n);
 }
@@ -650,6 +698,8 @@ char *path, *type;
 	else
 #endif
 	fp = fopen(path, type);
+	LOG2((*type == 'r') ? _LOG_INFO_ : _LOG_WARNING_, (fp) ? 0 : -1,
+		"fopen(\"%k\", \"%s\");", path, type);
 
 	return(fp);
 }
